# mypy: disallow_any_generics = False

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from io import StringIO
import re

import pandas as pd

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    ScanPositionSettingPlateReader,
)
from allotropy.allotrope.models.shared.definitions.definitions import JsonFloat
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    CalculatedDataItem,
    Data,
    DataSource,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.exceptions import (
    AllotropeConversionError,
)
from allotropy.parsers.agilent_gen5.constants import (
    DEFAULT_SOFTWARE_NAME,
    DEVICE_TYPE,
    EMISSION_KEY,
    EXCITATION_KEY,
    FILENAME_REGEX,
    GAIN_KEY,
    MEASUREMENTS_DATA_POINT_KEY,
    MIRROR_KEY,
    MULTIPLATE_FILE_ERROR,
    MULTIPLE_READ_MODE_ERROR,
    NAN_EMISSION_EXCITATION,
    OPTICS_KEY,
    PATHLENGTH_CORRECTION_KEY,
    READ_HEIGHT_KEY,
    READ_SPEED_KEY,
    ReadMode,
    ReadType,
    UNSUPPORTED_READ_MODE_ERROR,
    UNSUPPORTED_READ_TYPE_ERROR,
    WAVELENGTHS_KEY,
)
from allotropy.parsers.agilent_gen5.section_reader import SectionLinesReader
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.utils.pandas import df_to_series_data
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float,
    try_float_or_nan,
    try_float_or_none,
)


def read_data_section(reader: LinesReader) -> list[str]:
    data_section = [
        reader.pop() or "",
        *reader.pop_until_empty(),
    ]
    reader.drop_empty()
    return data_section


@dataclass(frozen=True)
class HeaderData:
    software_version: str
    experiment_file_path: str | None
    protocol_file_path: str | None
    datetime: str | None
    well_plate_identifier: str | None
    model_number: str | None
    equipment_serial_number: str | None
    file_name: str

    @classmethod
    def create(cls, reader: LinesReader, file_name: str) -> HeaderData:
        assert_not_none(reader.drop_until("^Software Version"), "Software Version")
        lines = [line for line in reader.pop_until("Procedure Details") if line]
        df = pd.read_csv(
            StringIO("\n".join(lines)),
            header=None,
            index_col=0,
            keep_default_na=False,
            sep="\t",
        ).T
        data = df_to_series_data(df, "Failed to parser header data")
        matches = re.match(FILENAME_REGEX, file_name)
        plate_identifier = matches.groupdict()["plate_identifier"] if matches else None
        date = data.get(str, "Date")
        time = data.get(str, "Time")
        return HeaderData(
            software_version=data[str, "Software Version"],
            experiment_file_path=data.get(str, "Experiment File Path:"),
            file_name=file_name,
            protocol_file_path=data.get(str, "Protocol File Path:"),
            datetime=f"{date} {time}" if date and time else None,
            well_plate_identifier=plate_identifier or data.get(str, "Plate Number"),
            model_number=data.get(str, "Reader Type:"),
            equipment_serial_number=data.get(str, "Reader Serial Number:"),
        )


@dataclass(frozen=True)
class FilterSet:
    gain: str
    emission: str | None = None
    excitation: str | None = None
    mirror: str | None = None
    optics: str | None = None

    @property
    def detector_wavelength_setting(self) -> JsonFloat | None:
        if not self.emission:
            return None
        return try_float_or_nan(self.emission.split("/")[0])

    @property
    def detector_bandwidth_setting(self) -> JsonFloat | None:
        if not self.emission:
            return None
        try:
            return try_float_or_nan(self.emission.split("/")[1])
        except IndexError:
            return None

    @property
    def excitation_wavelength_setting(self) -> JsonFloat | None:
        if self.excitation:
            return try_float_or_nan(self.excitation.split("/")[0])
        return None

    @property
    def excitation_bandwidth_setting(self) -> JsonFloat | None:
        if not self.excitation:
            return None
        try:
            return try_float(self.excitation.split("/")[1], "Excitation bandwith")
        except IndexError:
            return None

    @property
    def wavelength_filter_cutoff_setting(self) -> float | None:
        if self.mirror:
            return try_float(self.mirror.split(" ")[1], "Wavelength filter cutoff")
        return None

    @property
    def scan_position_setting(self) -> ScanPositionSettingPlateReader | None:
        position_lookup = {
            "Top": ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
            "Bottom": ScanPositionSettingPlateReader.bottom_scan_position__plate_reader_,
        }
        position = self.optics
        if self.mirror:
            position = self.mirror.split(" ")[0]

        # TODO: Raise if position is not a valid lookup value
        return position_lookup.get(position) if position is not None else None


@dataclass(frozen=True)
class ReadData:
    read_mode: ReadMode
    measurement_labels: list[str]
    pathlength_correction: str | None
    step_label: str | None
    number_of_averages: float | None
    detector_distance: float | None
    detector_carriage_speed: str | None
    filter_sets: dict[str, FilterSet]

    @classmethod
    def create(cls, reader: LinesReader) -> ReadData:
        assert_not_none(reader.drop_until("^Procedure Details"), "Procedure Details")
        reader.pop()
        reader.drop_empty()
        procedure_lines = read_data_section(reader)
        procedure_details = "\n".join(procedure_lines)
        read_type = cls.get_read_type(procedure_details)
        if read_type != ReadType.ENDPOINT:
            raise AllotropeConversionError(UNSUPPORTED_READ_TYPE_ERROR)

        read_mode = cls.get_read_mode(procedure_details)
        device_control_data = cls._get_device_control_data(procedure_details, read_mode)
        measurement_labels = cls._get_measurement_labels(device_control_data, read_mode)

        number_of_averages = device_control_data.get(MEASUREMENTS_DATA_POINT_KEY)
        read_height = device_control_data.get(READ_HEIGHT_KEY, "")

        return ReadData(
            read_mode=read_mode,
            step_label=device_control_data.get("Step Label"),
            measurement_labels=measurement_labels,
            detector_carriage_speed=device_control_data.get(READ_SPEED_KEY),
            # Absorbance attributes
            pathlength_correction=device_control_data.get(PATHLENGTH_CORRECTION_KEY),
            number_of_averages=try_float_or_none(number_of_averages),
            # Luminescence attributes
            detector_distance=try_float_or_none(read_height.split(" ")[0]),
            # Fluorescence attributes
            filter_sets=cls._get_filter_sets(
                measurement_labels, device_control_data, read_mode
            ),
        )

    @staticmethod
    def get_read_mode(procedure_details: str) -> ReadMode:
        read_modes = [
            read_mode for read_mode in ReadMode if read_mode.value in procedure_details
        ]
        if not read_modes:
            raise AllotropeConversionError(UNSUPPORTED_READ_MODE_ERROR)
        if len(read_modes) > 1:
            raise AllotropeConversionError(MULTIPLE_READ_MODE_ERROR)

        if read_modes[0] in (ReadMode.FLUORESCENCE, ReadMode.ALPHALISA):
            return ReadMode.FLUORESCENCE
        return read_modes[0]

    @staticmethod
    def get_read_type(procedure_details: str) -> ReadType:
        if ReadType.KINETIC.value in procedure_details:
            return ReadType.KINETIC
        elif ReadType.AREASCAN.value in procedure_details:
            return ReadType.AREASCAN
        elif ReadType.SPECTRAL.value in procedure_details:
            return ReadType.SPECTRAL
        # check for this last, because other modes still contain the word "Endpoint"
        elif ReadType.ENDPOINT.value in procedure_details:
            return ReadType.ENDPOINT

        msg = f"Read type not found; expected to find one of {sorted(ReadType._member_names_)}."
        raise AllotropeConversionError(msg)

    @classmethod
    def _get_measurement_labels(cls, device_control_data: dict, read_mode: str) -> list:
        step_label = device_control_data.get("Step Label")
        label_prefix = f"{step_label}:" if step_label else ""
        measurement_labels = []

        if read_mode == ReadMode.ABSORBANCE:
            measurement_labels = cls._get_absorbance_measurement_labels(
                label_prefix, device_control_data
            )

        if read_mode == ReadMode.FLUORESCENCE:
            excitations = device_control_data.get(EXCITATION_KEY, [])
            emissions = device_control_data.get(EMISSION_KEY, [])
            measurement_labels = [
                f"{label_prefix}{excitation},{emission}"
                for excitation, emission in zip(excitations, emissions, strict=True)
            ]
            if not measurement_labels:
                measurement_labels = ["Alpha"]

        if read_mode == ReadMode.LUMINESCENCE:
            emissions = device_control_data.get(EMISSION_KEY)
            for emission in emissions:
                label = "Lum" if emission in NAN_EMISSION_EXCITATION else emission
                measurement_labels.append(f"{label_prefix}{label}")

        return measurement_labels

    @classmethod
    def _get_absorbance_measurement_labels(
        cls, label_prefix: str | None, device_control_data: dict
    ) -> list:
        wavelengths = device_control_data.get(WAVELENGTHS_KEY, [])
        pathlength_correction = device_control_data.get(PATHLENGTH_CORRECTION_KEY)
        measurement_labels = []

        for wavelength in wavelengths:
            label = f"{label_prefix}{wavelength}"
            measurement_labels.append(label)

        if pathlength_correction:
            test, ref = pathlength_correction.split(" / ")
            test_label = f"{label_prefix}{test} [Test]"
            ref_label = f"{label_prefix}{ref} [Ref]"
            measurement_labels.extend([test_label, ref_label])

        return measurement_labels

    @classmethod
    def _get_device_control_data(
        cls, procedure_details: str, read_mode: ReadMode
    ) -> dict:
        list_keys = frozenset(
            {
                EMISSION_KEY,
                EXCITATION_KEY,
                OPTICS_KEY,
                GAIN_KEY,
                MIRROR_KEY,
                WAVELENGTHS_KEY,
            }
        )
        read_data_dict: dict = {label: [] for label in list_keys}
        read_lines: list[str] = procedure_details.splitlines()

        for line in read_lines:
            strp_line = str(line.strip())
            if strp_line.startswith("Read\t"):
                read_data_dict["Step Label"] = cls._get_step_label(line, read_mode)
                continue

            elif strp_line.startswith(WAVELENGTHS_KEY):
                wavelengths = strp_line.split(":  ")
                read_data_dict[WAVELENGTHS_KEY].extend(wavelengths[1].split(", "))
                continue

            line_data: list[str] = strp_line.split(",  ")
            for read_datum in line_data:
                splitted_datum = read_datum.split(": ")
                if len(splitted_datum) != 2:  # noqa: PLR2004
                    continue
                if splitted_datum[0] in list_keys:
                    read_data_dict[splitted_datum[0]].append(splitted_datum[1])
                else:
                    read_data_dict[splitted_datum[0]] = splitted_datum[1]

        return read_data_dict

    @classmethod
    def _get_step_label(cls, read_line: str, read_mode: str) -> str | None:
        split_line = read_line.strip().split("\t")
        if len(split_line) != 2:  # noqa: PLR2004
            msg = (
                f"Expected the Read data line {split_line} to contain exactly 2 values."
            )
            raise AllotropeConversionError(msg)
        if split_line[1] != f"{read_mode.title()} Endpoint":
            return split_line[1]

        return None

    @classmethod
    def _get_filter_sets(
        cls,
        measurement_labels: list[str],
        device_control_data: dict,
        read_mode: ReadMode,
    ) -> dict[str, FilterSet]:
        filter_data: dict[str, FilterSet] = {}
        if read_mode == ReadMode.ABSORBANCE:
            return filter_data

        emissions = device_control_data.get(EMISSION_KEY, [])
        excitations = device_control_data.get(EXCITATION_KEY, [])
        mirrors = device_control_data.get(MIRROR_KEY, [])
        optics = device_control_data.get(OPTICS_KEY, [])
        gains = device_control_data.get(GAIN_KEY, [])

        for idx, label in enumerate(measurement_labels):
            mirror = None
            if mirrors and read_mode == ReadMode.FLUORESCENCE:
                mirror = mirrors[idx]
            filter_data[label] = FilterSet(
                emission=emissions[idx] if emissions else None,
                gain=gains[idx],
                excitation=excitations[idx] if excitations else None,
                mirror=mirror,
                optics=optics[idx] if optics else None,
            )
        return filter_data


def get_identifiers(layout_lines: list[str] | None) -> dict[str, str]:
    if not layout_lines:
        return {}
    # Create dataframe from tabular data and forward fill empty values in index
    data = pd.read_csv(StringIO("\n".join(layout_lines[1:])), sep="\t")
    data = data.set_index(data.index.to_series().ffill(axis=0).values)

    identifiers = {}
    for row_name, row in data.iterrows():
        label = row.iloc[-1]
        for col_index, col in enumerate(row.iloc[:-1]):
            well_pos = f"{row_name}{col_index + 1}"
            # Prefer Name to Well ID
            if not pd.isna(col) and (
                label == "Name" or label == "Well ID" and well_pos not in identifiers
            ):
                identifiers[well_pos] = col
    return identifiers


def get_temperature(actual_temperature_lines: list[str] | None) -> float | None:
    if not actual_temperature_lines:
        return None
    if len(actual_temperature_lines) != 1:
        msg = f"Expected the Temperature section '{actual_temperature_lines}' to contain exactly 1 line."
        raise AllotropeConversionError(msg)

    return try_float(
        actual_temperature_lines[0].strip().split("\t")[-1], "Actual Temperature"
    )


@dataclass(frozen=True)
class MeasurementData:
    identifier: str
    value: JsonFloat
    label: str


def create_results(
    result_lines: list[str],
    header_data: HeaderData,
    read_data: ReadData,
    sample_identifiers: dict[str, str],
    actual_temperature: float | None,
) -> tuple[list[MeasurementGroup], list[CalculatedDataItem]]:
    if result_lines[0].strip() != "Results":
        msg = f"Expected the first line of the results section '{result_lines[0]}' to be 'Results'."
        raise AllotropeConversionError(msg)

    # Create dataframe from tabular data and forward fill empty values in index
    data = pd.read_csv(StringIO("\n".join(result_lines[1:])), sep="\t")
    data = data.set_index(data.index.to_series().ffill(axis=0).values)

    well_to_measurements: defaultdict[str, list[MeasurementData]] = defaultdict(list)
    calculated_data: defaultdict[str, list] = defaultdict(list)
    for row_name, row in data.iterrows():
        label = row.iloc[-1]
        for col_index, value in enumerate(row.iloc[:-1]):
            well_pos = f"{row_name}{col_index + 1}"
            well_value = try_float_or_nan(value)

            if label in read_data.measurement_labels:
                well_to_measurements[well_pos].append(
                    MeasurementData(random_uuid_str(), well_value, label)
                )
            else:
                calculated_data[well_pos].append((label, well_value))

    groups = [
        MeasurementGroup(
            plate_well_count=len(well_to_measurements),
            analytical_method_identifier=header_data.protocol_file_path,
            experimental_data_identifier=header_data.experiment_file_path,
            measurements=[
                _create_measurement(
                    measurement,
                    well_position,
                    header_data,
                    read_data,
                    sample_identifiers.get(well_position),
                    actual_temperature,
                )
                for measurement in measurements
            ],
        )
        for well_position, measurements in well_to_measurements.items()
    ]

    calculated_data_items = [
        CalculatedDataItem(
            identifier=random_uuid_str(),
            data_sources=[
                DataSource(
                    identifier=measurement.identifier,
                    feature=read_data.read_mode.value.lower(),
                )
                for measurement in _get_sources(
                    label, well_to_measurements[well_position]
                )
            ],
            unit=UNITLESS,
            name=label,
            value=value,
        )
        for well_position, well_calculated_data in calculated_data.items()
        for label, value in well_calculated_data
    ]

    return groups, calculated_data_items


def _get_sources(
    calculated_data_label: str, measurements: list[MeasurementData]
) -> list[MeasurementData]:
    # Pathlength is a special case, its sources are always determined
    # by the pathlength correction setting
    if calculated_data_label.split(":")[-1] == "Pathlength":
        pathlength_suffixes = ["[Test]", "[Ref]"]
        sources = [
            measurement
            for measurement in measurements
            if measurement.label.split(" ")[-1] in pathlength_suffixes
        ]
    else:
        sources = [
            measurement
            for measurement in measurements
            if measurement.label in calculated_data_label
        ]

    # if there are no matches in the measurement labels, use all measurements as source
    return sources or measurements


def _create_metadata(header_data: HeaderData, read_data: ReadData) -> Metadata:
    return Metadata(
        device_type=DEVICE_TYPE,
        detection_type=read_data.read_mode.value,
        device_identifier=NOT_APPLICABLE,
        model_number=header_data.model_number or NOT_APPLICABLE,
        equipment_serial_number=header_data.equipment_serial_number,
        software_name=DEFAULT_SOFTWARE_NAME,
        software_version=header_data.software_version,
        file_name=header_data.file_name,
        measurement_time=header_data.datetime,
    )


def _create_measurement(
    measurement: MeasurementData,
    well_position: str,
    header_data: HeaderData,
    read_data: ReadData,
    sample_identifier: str | None,
    actual_temperature: float | None,
) -> Measurement:
    # TODO(switch-statement): use switch statement once Benchling can use 3.10 syntax
    if read_data.read_mode == ReadMode.ABSORBANCE:
        measurement_type = MeasurementType.ULTRAVIOLET_ABSORBANCE
    elif read_data.read_mode == ReadMode.FLUORESCENCE:
        measurement_type = MeasurementType.FLUORESCENCE
    elif read_data.read_mode == ReadMode.LUMINESCENCE:
        measurement_type = MeasurementType.LUMINESCENCE

    detector_wavelength_setting: JsonFloat | None = None
    if measurement_type is MeasurementType.ULTRAVIOLET_ABSORBANCE:
        filter_data = None
        detector_wavelength_setting = float(
            measurement.label.split(":")[-1].split(" ")[0]
        )
    else:
        filter_data = read_data.filter_sets[measurement.label]
        detector_wavelength_setting = filter_data.detector_wavelength_setting

    return Measurement(
        type_=measurement_type,
        identifier=measurement.identifier,
        sample_identifier=sample_identifier
        or f"{header_data.well_plate_identifier} {well_position}",
        location_identifier=well_position,
        well_plate_identifier=header_data.well_plate_identifier,
        detector_wavelength_setting=detector_wavelength_setting,
        detector_bandwidth_setting=filter_data.detector_bandwidth_setting
        if filter_data
        else None,
        excitation_wavelength_setting=filter_data.excitation_wavelength_setting
        if filter_data
        else None,
        excitation_bandwidth_setting=filter_data.excitation_bandwidth_setting
        if filter_data
        else None,
        wavelength_filter_cutoff_setting=filter_data.wavelength_filter_cutoff_setting
        if filter_data
        else None,
        detector_distance_setting=read_data.detector_distance,
        scan_position_setting=filter_data.scan_position_setting
        if filter_data
        else None,
        detector_gain_setting=filter_data.gain if filter_data else None,
        number_of_averages=read_data.number_of_averages,
        detector_carriage_speed=read_data.detector_carriage_speed,
        absorbance=measurement.value
        if measurement_type == MeasurementType.ULTRAVIOLET_ABSORBANCE
        else None,
        fluorescence=measurement.value
        if measurement_type == MeasurementType.FLUORESCENCE
        else None,
        luminescence=measurement.value
        if measurement_type == MeasurementType.LUMINESCENCE
        else None,
        compartment_temperature=actual_temperature,
    )


def create_data(reader: SectionLinesReader, file_name: str) -> Data:
    plates = list(reader.iter_sections("^Software Version"))

    if not plates:
        msg = "No plate data found in file."
        raise AllotropeConversionError(msg)

    if len(plates) > 1:
        raise AllotropeConversionError(MULTIPLATE_FILE_ERROR)

    header_data = HeaderData.create(plates[0], file_name)
    read_data = ReadData.create(plates[0])

    section_lines = {}
    while plates[0].current_line_exists():
        data_section = read_data_section(plates[0])
        section_lines[data_section[0].strip().split(":")[0]] = data_section
    if "Results" not in section_lines:
        msg = "Missing 'Results' section"
        raise AllotropeConversionError(msg)

    sample_identifiers = get_identifiers(section_lines.get("Layout"))
    actual_temperature = get_temperature(section_lines.get("Actual Temperature"))

    measurement_groups, calculated_data = create_results(
        section_lines["Results"],
        header_data,
        read_data,
        sample_identifiers,
        actual_temperature,
    )

    return Data(
        metadata=_create_metadata(header_data, read_data),
        measurement_groups=measurement_groups,
        calculated_data=calculated_data,
    )



from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from typing import Any

from allotropy.allotrope.converter import add_custom_information_document
from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    CalculatedDataAggregateDocument,
    CalculatedDataDocumentItem,
    ContainerType,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceSystemDocument,
    FluorescencePointDetectionDeviceControlAggregateDocument,
    FluorescencePointDetectionDeviceControlDocumentItem,
    FluorescencePointDetectionMeasurementDocumentItems,
    MeasurementAggregateDocument,
    Model,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
    SpectrophotometryAggregateDocument,
    SpectrophotometryDocumentItem,
    UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument,
    UltravioletAbsorbancePointDetectionDeviceControlDocumentItem,
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMicroliter,
    TQuantityValueMilliAbsorbanceUnit,
    TQuantityValueNanometer,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    InvalidJsonFloat,
    JsonFloat,
    TDateTimeValue,
    TQuantityValue,
)
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.exceptions import AllotropyParserError
from allotropy.parsers.utils.units import get_quantity_class
from allotropy.parsers.utils.values import assert_not_none, quantity_or_none


class MeasurementType(Enum):
    ULTRAVIOLET_ABSORBANCE = "ULTRAVIOLET_ABSORBANCE"
    FLUORESCENCE = "FLUORESCENCE"


MeasurementDocumentItems = (
    UltravioletAbsorbancePointDetectionMeasurementDocumentItems
    | FluorescencePointDetectionMeasurementDocumentItems
)


@dataclass(frozen=True)
class ProcessedDataFeature:
    result: float | InvalidJsonFloat
    unit: str
    feature: str | None = None
    identifier: str | None = None


@dataclass(frozen=True)
class ProcessedData:
    features: list[ProcessedDataFeature]
    identifier: str | None = None


@dataclass(frozen=True)
class DataSource:
    identifier: str
    feature: str


@dataclass(frozen=True)
class CalculatedDataItem:
    identifier: str
    name: str
    value: float
    unit: str
    data_sources: list[DataSource]


@dataclass(frozen=True)
class Measurement:
    type_: MeasurementType
    # Measurement metadata
    identifier: str
    sample_identifier: str
    location_identifier: str | None = None
    batch_identifier: str | None = None
    analyst: str | None = None
    measurement_time: str | None = None
    well_plate_identifier: str | None = None

    # Settings
    sample_volume_setting: float | None = None
    detector_wavelength_setting: JsonFloat | None = None
    excitation_wavelength_setting: str | None = None
    emission_wavelength_setting: str | None = None
    dilution_factor_setting: float | None = None
    original_sample_concentration: JsonFloat | None = None
    original_sample_concentration_unit: str | None = None
    qubit_tube_concentration: JsonFloat | None = None
    qubit_tube_concentration_units: str | None = None
    standard_1_concentration: float | None = None
    standard_2_concentration: float | None = None
    standard_3_concentration: float | None = None

    # Measurements
    absorbance: JsonFloat | None = None
    fluorescence: JsonFloat | None = None

    # Processed data
    calculated_data: list[CalculatedDataItem] | None = None
    processed_data: ProcessedData | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurements: list[Measurement]
    plate_well_count: float | None = None
    measurement_time: str | None = None
    experiment_type: str | None = None
    analyst: str | None = None
    container_type: str | None = None

    processed_data: ProcessedData | None = None


@dataclass(frozen=True)
class Metadata:
    device_identifier: str
    device_type: str
    model_number: str
    software_name: str | None = None
    detection_type: str | None = None
    unc_path: str | None = None
    software_version: str | None = None
    equipment_serial_number: str | None = None
    product_manufacturer: str | None = None
    brand_name: str | None = None
    container_type: ContainerType | None = None

    file_name: str | None = None
    data_system_instance_id: str | None = None

    analyst: str | None = None
    measurement_time: str | None = None


@dataclass(frozen=True)
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]
    calculated_data: list[CalculatedDataItem] | None = None


class Mapper:
    MANIFEST = "http://purl.allotrope.org/manifests/spectrophotometry/BENCHLING/2023/12/spectrophotometry.manifest"

    def __init__(
        self, asm_converter_name: str, get_date_time: Callable[[str], TDateTimeValue]
    ) -> None:
        self.converter_name = asm_converter_name
        self.get_date_time = get_date_time

    def map_model(self, data: Data) -> Model:
        return Model(
            spectrophotometry_aggregate_document=SpectrophotometryAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    device_identifier=data.metadata.device_identifier,
                    model_number=data.metadata.model_number,
                    brand_name=data.metadata.brand_name,
                    product_manufacturer=data.metadata.product_manufacturer,
                ),
                data_system_document=DataSystemDocument(
                    file_name=data.metadata.file_name,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                    software_name=data.metadata.software_name,
                ),
                spectrophotometry_document=[
                    self._get_technique_document(measurement_group, data.metadata)
                    for measurement_group in data.measurement_groups
                ],
                calculated_data_aggregate_document=self._get_calculated_data_aggregate_document(
                    data.calculated_data
                ),
            ),
            field_asm_manifest=self.MANIFEST,
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> SpectrophotometryDocumentItem:
        return SpectrophotometryDocumentItem(
            analyst=measurement_group.analyst,
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_time=self.get_date_time(
                    assert_not_none(measurement_group.measurement_time)
                ),
                experiment_type=measurement_group.experiment_type,
                container_type=metadata.container_type,
                measurement_document=[
                    self._get_measurement_document_item(measurement, metadata)
                    for measurement in measurement_group.measurements
                ],
            ),
        )

    def _get_measurement_document_item(
        self, measurement: Measurement, metadata: Metadata
    ) -> MeasurementDocumentItems:
        # TODO(switch-statement): use switch statement once Benchling can use 3.10 syntax
        if measurement.type_ == MeasurementType.ULTRAVIOLET_ABSORBANCE:
            return self._get_ultraviolet_absorbance_measurement_document(
                measurement, metadata
            )
        elif measurement.type_ == MeasurementType.FLUORESCENCE:
            return self._get_fluorescence_measurement_document(measurement, metadata)
        else:
            msg = f"Invalid measurement type: {measurement.type}"
            raise AllotropyParserError(msg)

    def _get_ultraviolet_absorbance_measurement_document(
        self, measurement: Measurement, metadata: Metadata
    ) -> UltravioletAbsorbancePointDetectionMeasurementDocumentItems:
        return UltravioletAbsorbancePointDetectionMeasurementDocumentItems(
            measurement_identifier=measurement.identifier,
            sample_document=self._get_sample_document(measurement),
            processed_data_aggregate_document=self._get_processed_data_aggregate_document(
                measurement.processed_data
            ),
            device_control_aggregate_document=UltravioletAbsorbancePointDetectionDeviceControlAggregateDocument(
                device_control_document=[
                    add_custom_information_document(
                        UltravioletAbsorbancePointDetectionDeviceControlDocumentItem(
                            device_type=metadata.device_type,
                            detector_wavelength_setting=quantity_or_none(
                                TQuantityValueNanometer,
                                measurement.detector_wavelength_setting,
                            ),
                        ),
                        self._get_device_control_custom_document(measurement),
                    )
                ]
            ),
            absorbance=TQuantityValueMilliAbsorbanceUnit(
                value=assert_not_none(measurement.absorbance)  # type: ignore[arg-type]
            ),
            calculated_data_aggregate_document=self._get_calculated_data_aggregate_document(
                measurement.calculated_data
            ),
        )

    def _get_fluorescence_measurement_document(
        self, measurement: Measurement, metadata: Metadata
    ) -> FluorescencePointDetectionMeasurementDocumentItems:
        return FluorescencePointDetectionMeasurementDocumentItems(
            measurement_identifier=measurement.identifier,
            sample_document=self._get_sample_document(measurement),
            processed_data_aggregate_document=self._get_processed_data_aggregate_document(
                measurement.processed_data
            ),
            device_control_aggregate_document=FluorescencePointDetectionDeviceControlAggregateDocument(
                device_control_document=[
                    add_custom_information_document(
                        FluorescencePointDetectionDeviceControlDocumentItem(
                            device_type=metadata.device_type,
                            detector_wavelength_setting=quantity_or_none(
                                TQuantityValueNanometer,
                                measurement.detector_wavelength_setting,
                            ),
                        ),
                        self._get_device_control_custom_document(measurement),
                    )
                ]
            ),
            fluorescence=TQuantityValueRelativeFluorescenceUnit(
                value=assert_not_none(measurement.fluorescence)  # type: ignore[arg-type]
            ),
        )

    def _get_device_control_custom_document(
        self, measurement: Measurement
    ) -> dict[str, Any]:
        return {
            "sample volume setting": quantity_or_none(
                TQuantityValueMicroliter, measurement.sample_volume_setting
            ),
            "excitation setting": measurement.excitation_wavelength_setting,
            "emission setting": measurement.emission_wavelength_setting,
            "dilution factor": quantity_or_none(
                TQuantityValueUnitless, measurement.dilution_factor_setting
            ),
        }

    def _get_sample_document(self, measurement: Measurement) -> SampleDocument:
        custom_document = {
            "original sample concentration": quantity_or_none(
                get_quantity_class(measurement.original_sample_concentration_unit)
                or TQuantityValueUnitless,
                measurement.original_sample_concentration,
            ),
            "qubit tube concentration": quantity_or_none(
                get_quantity_class(measurement.qubit_tube_concentration_units)
                or TQuantityValueUnitless,
                measurement.qubit_tube_concentration,
            ),
            "standard 1 concentration": quantity_or_none(
                TQuantityValueRelativeFluorescenceUnit,
                measurement.standard_1_concentration,
            ),
            "standard 2 concentration": quantity_or_none(
                TQuantityValueRelativeFluorescenceUnit,
                measurement.standard_2_concentration,
            ),
            "standard 3 concentration": quantity_or_none(
                TQuantityValueRelativeFluorescenceUnit,
                measurement.standard_3_concentration,
            ),
        }
        return add_custom_information_document(
            SampleDocument(
                sample_identifier=measurement.sample_identifier,
                batch_identifier=measurement.batch_identifier,
                location_identifier=measurement.location_identifier,
                well_plate_identifier=measurement.well_plate_identifier,
            ),
            custom_document,
        )

    def _get_processed_data_aggregate_document(
        self, data: ProcessedData | None
    ) -> ProcessedDataAggregateDocument | None:
        if not data:
            return None

        return ProcessedDataAggregateDocument(
            processed_data_document=[
                ProcessedDataDocumentItem(
                    # TODO(nstender): figure out how to limit possible classes from get_quantity_class for typing.
                    mass_concentration=quantity_or_none(
                        get_quantity_class(feature.unit), feature.result  # type: ignore[arg-type]
                    ),
                    processed_data_identifier=data.identifier,
                )
                for feature in data.features
            ]
        )

    def _get_calculated_data_aggregate_document(
        self, calculated_data_items: list[CalculatedDataItem] | None
    ) -> CalculatedDataAggregateDocument | None:
        if not calculated_data_items:
            return None

        return CalculatedDataAggregateDocument(
            calculated_data_document=[
                CalculatedDataDocumentItem(
                    calculated_data_identifier=calculated_data_item.identifier,
                    calculated_data_name=calculated_data_item.name,
                    calculated_result=TQuantityValue(
                        value=calculated_data_item.value,
                        unit=calculated_data_item.unit,
                    ),
                    data_source_aggregate_document=DataSourceAggregateDocument(
                        data_source_document=[
                            DataSourceDocumentItem(
                                data_source_identifier=item.identifier,
                                data_source_feature=item.feature,
                            )
                            for item in calculated_data_item.data_sources
                        ]
                    ),
                )
                for calculated_data_item in calculated_data_items
            ]
        )



from collections.abc import Callable
from dataclasses import dataclass

from allotropy.allotrope.models.adm.cell_counting.benchling._2023._11.cell_counting import (
    CellCountingAggregateDocument,
    CellCountingDetectorDeviceControlAggregateDocument,
    CellCountingDetectorMeasurementDocumentItem,
    CellCountingDocumentItem,
    DataProcessingDocument,
    DataSystemDocument,
    DeviceControlDocumentItemModel,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    Model,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCell,
    TQuantityValueMicrometer,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValuePercent,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import TDateTimeValue
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.parsers.utils.values import quantity_or_none


@dataclass
class Measurement:
    measurement_identifier: str
    timestamp: str
    sample_identifier: str
    analyst: str
    viability: float
    viable_cell_density: float
    cell_type_processing_method: str | None = None
    minimum_cell_diameter_setting: float | None = None
    maximum_cell_diameter_setting: float | None = None
    cell_density_dilution_factor: float | None = None
    total_cell_count: float | None = None
    total_cell_density: float | None = None
    average_total_cell_diameter: float | None = None
    average_live_cell_diameter: float | None = None
    viable_cell_count: float | None = None
    average_total_cell_circularity: float | None = None
    average_viable_cell_circularity: float | None = None


@dataclass(frozen=True)
class MeasurementGroup:
    measurements: list[Measurement]
    _analyst: str | None = None

    @property
    def analyst(self) -> str | None:
        if self._analyst is not None:
            return self._analyst
        # If all measurements in the group have the same analyst set, set at the group level too.
        if (
            self.measurements
            and len({m.analyst for m in self.measurements}) == 1
            and self.measurements[0].analyst
        ):
            return self.measurements[0].analyst
        return None


@dataclass
class Metadata:
    device_type: str
    detection_type: str
    model_number: str
    software_name: str
    file_name: str
    equipment_serial_number: str | None = None
    software_version: str | None = None


@dataclass
class Data:
    metadata: Metadata
    measurement_groups: list[MeasurementGroup]


class Mapper:
    MANIFEST = "http://purl.allotrope.org/manifests/cell-counting/BENCHLING/2023/11/cell-counting.manifest"

    def __init__(
        self, asm_converter_name: str, get_date_time: Callable[[str], TDateTimeValue]
    ) -> None:
        self.converter_name = asm_converter_name
        self.get_date_time = get_date_time

    def map_model(self, data: Data) -> Model:
        return Model(
            field_asm_manifest=self.MANIFEST,
            cell_counting_aggregate_document=CellCountingAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    model_number=data.metadata.model_number,
                    equipment_serial_number=data.metadata.equipment_serial_number,
                ),
                data_system_document=DataSystemDocument(
                    file_name=data.metadata.file_name,
                    software_name=data.metadata.software_name,
                    software_version=data.metadata.software_version,
                    ASM_converter_name=self.converter_name,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                cell_counting_document=[
                    self._get_technique_document(measurement_group, data.metadata)
                    for measurement_group in data.measurement_groups
                ],
            ),
        )

    def _get_technique_document(
        self, measurement_group: MeasurementGroup, metadata: Metadata
    ) -> CellCountingDocumentItem:
        return CellCountingDocumentItem(
            analyst=measurement_group.analyst,
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_document=[
                    self._get_measurement_document(measurement, metadata)
                    for measurement in measurement_group.measurements
                ]
            ),
        )

    def _get_measurement_document(
        self, measurement: Measurement, metadata: Metadata
    ) -> CellCountingDetectorMeasurementDocumentItem:
        return CellCountingDetectorMeasurementDocumentItem(
            measurement_time=self.get_date_time(measurement.timestamp),
            measurement_identifier=measurement.measurement_identifier,
            sample_document=SampleDocument(
                sample_identifier=measurement.sample_identifier
            ),
            device_control_aggregate_document=CellCountingDetectorDeviceControlAggregateDocument(
                device_control_document=[
                    DeviceControlDocumentItemModel(
                        device_type=metadata.device_type,
                        detection_type=metadata.detection_type,
                    )
                ]
            ),
            processed_data_aggregate_document=ProcessedDataAggregateDocument(
                processed_data_document=[
                    ProcessedDataDocumentItem(
                        data_processing_document=DataProcessingDocument(
                            cell_type_processing_method=measurement.cell_type_processing_method,
                            minimum_cell_diameter_setting=quantity_or_none(
                                TQuantityValueMicrometer,
                                measurement.minimum_cell_diameter_setting,
                            ),
                            maximum_cell_diameter_setting=quantity_or_none(
                                TQuantityValueMicrometer,
                                measurement.maximum_cell_diameter_setting,
                            ),
                            cell_density_dilution_factor=quantity_or_none(
                                TQuantityValueUnitless,
                                measurement.cell_density_dilution_factor,
                            ),
                        ),
                        viability__cell_counter_=TQuantityValuePercent(
                            value=measurement.viability
                        ),
                        viable_cell_density__cell_counter_=TQuantityValueMillionCellsPerMilliliter(
                            value=measurement.viable_cell_density
                        ),
                        total_cell_count=quantity_or_none(
                            TQuantityValueCell, measurement.total_cell_count
                        ),
                        total_cell_density__cell_counter_=quantity_or_none(
                            TQuantityValueMillionCellsPerMilliliter,
                            measurement.total_cell_density,
                        ),
                        average_total_cell_diameter=quantity_or_none(
                            TQuantityValueMicrometer,
                            measurement.average_total_cell_diameter,
                        ),
                        average_live_cell_diameter__cell_counter_=quantity_or_none(
                            TQuantityValueMicrometer,
                            measurement.average_live_cell_diameter,
                        ),
                        viable_cell_count=quantity_or_none(
                            TQuantityValueCell, measurement.viable_cell_count
                        ),
                        average_total_cell_circularity=quantity_or_none(
                            TQuantityValueUnitless,
                            measurement.average_total_cell_circularity,
                        ),
                        average_viable_cell_circularity=quantity_or_none(
                            TQuantityValueUnitless,
                            measurement.average_viable_cell_circularity,
                        ),
                    ),
                ]
            ),
        )



"""Reader file for Roche Cedex HiRes Instrument"""

import pandas as pd

from allotropy.allotrope.pandas_util import read_csv, read_excel
from allotropy.constants import DEFAULT_ENCODING
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.roche_cedex_hires import constants


class RocheCedexHiResReader:
    """
    A reader class for parsing Roche Cedex HiRes data files into pandas DataFrames.

    Methods:
    read(named_file_contents: NamedFileContents) -> pd.DataFrame:
        Reads the content of the provided named file and returns it as a pandas DataFrame.
    """

    @classmethod
    def read(cls, named_file_contents: NamedFileContents) -> pd.DataFrame:
        """
        Reads the content of the provided named file and returns it as a pandas DataFrame.

        Parameters:
        named_file_contents (NamedFileContents): The named file contents to read.
            It includes the original file name, file contents, and encoding.

        Returns:
        pd.DataFrame: The content of the file as a pandas DataFrame.

        Raises:
        AllotropeConversionError: If the file format is not supported.
        """
        if named_file_contents.original_file_name.endswith(".csv"):
            return read_csv(
                named_file_contents.contents,
                index_col=False,
                encoding=DEFAULT_ENCODING,
            )
        elif named_file_contents.original_file_name.endswith(".xlsx"):
            return read_excel(named_file_contents.contents.name)
        else:
            message = f"{constants.UNSUPPORTED_FILE_FORMAT_ERROR} '{named_file_contents.original_file_name}'"
            raise AllotropeConversionError(message)



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum
import logging
import re

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    ScanPositionSettingPlateReader,
    TransmittedLightSetting,
)
from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.exceptions import AllotropeConversionError, AllotropeParsingError
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float,
    try_float_or_none,
)


class ExperimentType(Enum):
    FLUORESCENCE = "fluorescence"
    ABSORBANCE = "absorbance"
    LUMINESCENCE = "luminescence"
    OPTICAL_IMAGING = "optical imaging"


class Version(Enum):
    V2 = "2.0"
    V3 = "3.0"


PLATEMAP_TO_SAMPLE_ROLE_TYPE = {
    "B": SampleRoleType.blank_role.value,
    "C": SampleRoleType.control_sample_role.value,
    "S": SampleRoleType.standard_sample_role.value,
    "U": SampleRoleType.unknown_sample_role.value,
    "E": SampleRoleType.control_sample_role.value,
    "ZL": SampleRoleType.control_sample_role.value,
    "ZH": SampleRoleType.control_sample_role.value,
    "LB": SampleRoleType.control_sample_role.value,
    "LC": SampleRoleType.control_sample_role.value,
    "LH": SampleRoleType.control_sample_role.value,
}


SCAN_POSITION_CONVERSION = {
    "TOP": ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
    "BOTTOM": ScanPositionSettingPlateReader.bottom_scan_position__plate_reader_,
}


TRANSMITTED_LIGHT_CONVERSION = {
    "BRIGHTFIELD": TransmittedLightSetting.brightfield,
    "DARKFIELD": TransmittedLightSetting.darkfield,
    "PHASE CONTRAST": TransmittedLightSetting.phase_contrast,
}


@dataclass
class BackgroundInfo:
    experiment_type: str


@dataclass(frozen=True)
class Results:
    barcode: str
    data: dict[str, str]

    def get_plate_well_count(self) -> int:
        return len(self.data)

    def get_well_float_value(self, well_position: str) -> float:
        return try_float(
            self.get_well_str_value(well_position),
            f"result well at '{well_position}'",
        )

    def get_well_str_value(self, well_position: str) -> str:
        return str(
            assert_not_none(
                self.data.get(well_position),
                msg=f"Unable to get well at position '{well_position}' from results section.",
            )
        )


@dataclass(frozen=True)
class AnalysisResult:
    analysis_parameter: str
    results: dict[str, str]

    @staticmethod
    def create(reader: CsvReader) -> AnalysisResult | None:
        barcode_line = assert_not_none(
            reader.drop_until_inclusive("^Barcode:(.+),"),
            msg="Unable to find background information.",
        )

        analysis_parameter = None
        for element in barcode_line.split(","):
            if re.search("^.+:.+$", element):
                key, value = element.split(":", maxsplit=1)
                if "AnalysisParameter" in key:
                    analysis_parameter = value
                    break

        analysis_parameter = assert_not_none(
            analysis_parameter,
            msg="Unable to find analysis parameter in Analysis Results section.",
        )

        try:
            results_df = assert_not_none(
                reader.pop_csv_block_as_df(header=0, index_col=0),
                msg="Unable to find results table.",
            )
        except AllotropeParsingError:
            logging.warning(
                f"Unable to read analysis result '{analysis_parameter}'. Ignoring"
            )
            return None

        for column in results_df:
            if str(column).startswith("Unnamed"):
                results_df = results_df.drop(columns=column)

        results = {
            f"{row}{col}": values[col]
            for row, values in results_df.iterrows()
            for col in results_df.columns
        }

        a1_value = str(
            assert_not_none(
                results.get("A1"),
                msg=f"Unable to find first position of analysis result '{analysis_parameter}'.",
            )
        )

        # if first value is not a valid float value the results are not useful
        if try_float_or_none(a1_value) is None:
            return None

        return AnalysisResult(
            analysis_parameter=analysis_parameter,
            results=results,
        )

    def get_result(self, well_position: str) -> str:
        return str(
            assert_not_none(
                self.results.get(str(well_position)),
                msg=f"Unable to get well at position '{well_position}' from analysis result '{self.analysis_parameter}'.",
            )
        )

    def get_image_feature_result(self, well_position: str) -> float:
        return try_float(
            self.get_result(well_position),
            f"analysis result '{self.analysis_parameter}' at '{well_position}'",
        )


@dataclass(frozen=True)
class MeasurementInfo:
    instrument_serial_number: str
    measurement_time: str
    protocol_signature: str
    measurement_signature: str

    @staticmethod
    def create(elements: dict[str, str]) -> MeasurementInfo:
        instrument_serial_number = assert_not_none(
            elements.get("Instrument Serial Number"),
            msg="Unable to find Instrument Serial Number in Measurement Information section.",
        )

        measurement_time = assert_not_none(
            elements.get("Measurement Started"),
            msg="Unable to find Measurement time in Measurement Information section.",
        )

        protocol_signature = assert_not_none(
            elements.get("Protocol Signature"),
            msg="Unable to find Protocol Signature in Measurement Information section.",
        )

        measurement_signature = assert_not_none(
            elements.get("Measurement Signature"),
            msg="Unable to find Measurement Signature in Measurement Information section.",
        )

        return MeasurementInfo(
            instrument_serial_number,
            measurement_time,
            protocol_signature,
            measurement_signature,
        )


@dataclass(frozen=True)
class Platemap:
    data: dict[str, str]

    def get_well_value(self, well_position: str) -> str:
        return str(
            assert_not_none(
                self.data.get(well_position),
                msg=f"Unable to get well at position '{well_position}' from platemap section.",
            )
        )

    def get_sample_role_type(self, well_position: str) -> str | None:
        raw_value = self.get_well_value(well_position)
        if raw_value == "-":
            return None

        value = assert_not_none(
            re.match(r"^([A-Z]+)\d+$", raw_value),
            msg=f"Unable to understand platemap value '{raw_value}' for well position '{well_position}'.",
        ).group(1)

        return assert_not_none(
            PLATEMAP_TO_SAMPLE_ROLE_TYPE.get(value),
            msg=f"Unable to find sample role type for well position '{well_position}'.",
        )


@dataclass(frozen=True)
class MeasurementElement:
    title: str
    value: str


@dataclass(frozen=True)
class Channel:
    name: str
    fluorescent_tag: str | None
    transmitted_light: TransmittedLightSetting | None
    excitation_wavelength: float
    illumination: float
    exposure_duration: float
    additional_focus_offset: float

    @staticmethod
    def check_element_title(element: MeasurementElement, expected_title: str) -> None:
        if element.title != expected_title:
            msg = f"Expected to get '{expected_title}' but '{element.title}' was found."
            raise AllotropeConversionError(msg)

    @staticmethod
    def create(
        name: MeasurementElement,
        excitation_wavelength: MeasurementElement,
        excitation_power: MeasurementElement,
        exposure_time: MeasurementElement,
        additional_focus_offset: MeasurementElement,
    ) -> Channel:
        Channel.check_element_title(name, "Channel")
        Channel.check_element_title(excitation_wavelength, "Excitation wavelength [nm]")
        Channel.check_element_title(excitation_power, "Excitation Power [%]")
        Channel.check_element_title(exposure_time, "Exposure Time [ms]")
        Channel.check_element_title(
            additional_focus_offset, "Additional Focus offset [mm]"
        )

        return Channel(
            name.value,
            fluorescent_tag=None if name.value == "BRIGHTFIELD" else name.value,
            transmitted_light=TRANSMITTED_LIGHT_CONVERSION.get(name.value),
            excitation_wavelength=try_float(
                excitation_wavelength.value.removesuffix("nm"), "excitation wavelength"
            ),
            illumination=try_float(excitation_power.value, "excitation power"),
            exposure_duration=try_float(exposure_time.value, "exposure time"),
            additional_focus_offset=try_float(
                additional_focus_offset.value, "additional focus offset"
            ),
        )


@dataclass(frozen=True)
class Measurements:
    channels: list[Channel]
    number_of_averages: float | None
    detector_distance: float | None
    scan_position: ScanPositionSettingPlateReader | None
    emission_wavelength: float | None
    excitation_wavelength: float | None
    focus_height: float | None

    @staticmethod
    def create_channels(elements: list[MeasurementElement]) -> list[Channel]:
        channels: list[Channel] = []
        for idx, element in enumerate(elements):
            if element.title != "Channel":
                continue
            try:
                channels.append(
                    Channel.create(
                        name=elements[idx],
                        excitation_wavelength=elements[idx + 1],
                        excitation_power=elements[idx + 2],
                        exposure_time=elements[idx + 3],
                        additional_focus_offset=elements[idx + 4],
                    )
                )
            except IndexError as e:
                msg = f"Unable to get channel data for channel {element.value}"
                raise AllotropeConversionError(msg) from e
        return channels

    @staticmethod
    def try_element_or_none(
        elements: list[MeasurementElement], title: str
    ) -> MeasurementElement | None:
        for element in elements:
            if element.title == title:
                return element
        return None

    @staticmethod
    def get_element_float_value_or_none(
        elements: list[MeasurementElement], title: str
    ) -> float | None:
        element = Measurements.try_element_or_none(elements, title)
        if element is None:
            return None
        return try_float_or_none(element.value)


@dataclass(frozen=True)
class Data:
    version: str
    background_info: BackgroundInfo
    results: Results
    analysis_results: list[AnalysisResult]
    measurement_info: MeasurementInfo
    platemap: Platemap
    measurements: Measurements

    def iter_wells(self) -> Iterator[str]:
        yield from self.results.data

    def get_well_float_value(self, well_position: str) -> float:
        return self.results.get_well_float_value(well_position)

    def get_well_str_value(self, well_position: str) -> str:
        return self.results.get_well_str_value(well_position)

    def get_platemap_well_value(self, well_position: str) -> str:
        return self.platemap.get_well_value(well_position)

    def get_sample_role_type(self, well_position: str) -> str | None:
        return self.platemap.get_sample_role_type(well_position)



import pandas as pd

from allotropy.allotrope.pandas_util import read_csv
from allotropy.types import IOType


class NucleoviewReader:
    @classmethod
    def read(cls, contents: IOType) -> pd.DataFrame:
        df = read_csv(contents, sep=";", skipinitialspace=True, index_col=0)
        df = df[:-1].dropna(axis=0, how="all").T
        df = df.rename(
            {"Estimated cell diameter [um]": "Estimated cell diameter (um)"},
            axis="columns",
        )
        return df



from datetime import timedelta, timezone, tzinfo
from zoneinfo import ZoneInfo

from dateutil import parser
import pytz

from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.utils.values import assert_not_none

TIMEZONE_CODES_MAP = {
    **{code: pytz.timezone(code) for code in pytz.all_timezones},
    # Add daylight savings time codes for USA
    **{
        "EDT": timezone(timedelta(hours=-4), "EDT"),
        "CDT": timezone(timedelta(hours=-5), "CDT"),
        "MDT": timezone(timedelta(hours=-6), "MDT"),
        "PDT": timezone(timedelta(hours=-7), "PDT"),
    },
}


# TODO: TimestampParser should support localization -- e.g., passing "dayfirst=True" to dateutil.parser.parse.
class TimestampParser:
    default_timezone: tzinfo

    def __init__(self, default_timezone: tzinfo | None = None):
        if default_timezone and not isinstance(default_timezone, tzinfo):
            msg = f"Invalid default timezone '{default_timezone}'."
            raise AllotropeConversionError(msg)
        self.default_timezone = default_timezone or ZoneInfo("UTC")

    def parse(self, time: str) -> str:
        """Parse a string to a datetime, then format as an ISO 8601 string.

        If the parsed datetime doesn't have a timezone, use self.default_timezone.

        :param time: the string to parse
        :raises AllotropeConversionError if time cannot be parsed
        """
        assert_not_none(time, "time")

        try:
            timestamp = parser.parse(time, tzinfos=TIMEZONE_CODES_MAP, fuzzy=True)
        except ValueError as e:
            msg = f"Could not parse time '{time}'."
            raise AllotropeConversionError(msg) from e
        if not timestamp.tzinfo:
            timestamp = timestamp.replace(tzinfo=self.default_timezone)
        return str(timestamp.isoformat())



# plate (repeated for N plates measured)
#     Plate information
#     Background information (optional)
#     Calculated results (optional)
#     Measured results (repeated for N measurements performed per plate)
# Basic assay information
# Protocol information
# Plate type
# Platemap (repeated for N plates measured)
# Calculations
# Auto export parameters
# Operations
# Labels
#     Filters (dependent on detection modality)
#     Mirrors (dependent on detection modality)
# Instrument
from __future__ import annotations

from dataclasses import dataclass
from re import search
from typing import Any

import numpy as np
import pandas as pd

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    ScanPositionSettingPlateReader,
)
from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    num_to_chars,
)


# TODO(nstender): figure out how to integrate this with pandas util function
def df_to_series(df: pd.DataFrame) -> pd.Series[Any]:
    df.columns = df.iloc[0]  # type: ignore[assignment]
    return pd.Series(df.iloc[-1], index=df.columns)


@dataclass(frozen=True)
class PlateInfo:
    number: str
    barcode: str
    measurement_time: str | None
    measured_height: float | None
    chamber_temperature_at_start: float | None

    @staticmethod
    def get_series(reader: CsvReader) -> SeriesData:
        assert_not_none(
            reader.pop_if_match("^Plate information"),
            msg="Unable to find expected plate information",
        )

        return SeriesData(
            df_to_series(
                assert_not_none(
                    reader.pop_csv_block_as_df(),
                    "Plate information CSV block",
                )
            ).replace(np.nan, None)
        )


@dataclass(frozen=True)
class CalculatedPlateInfo(PlateInfo):
    formula: str
    name: str

    @staticmethod
    def create(data: SeriesData) -> CalculatedPlateInfo:
        plate_number = data[str, "Plate"]
        formula = data[
            str,
            "Formula",
            "Unable to find expected formula for calculated results section.",
        ]

        name = assert_not_none(
            search(r"^([^=]*)=", formula),
            msg="Unable to find expected formula name for calculated results section.",
        ).group(1)

        raw_barcode = data.get(str, "Barcode")
        barcode = (raw_barcode or '=""').removeprefix('="').removesuffix('"')
        barcode = barcode or f"Plate {plate_number}"

        return CalculatedPlateInfo(
            number=plate_number,
            barcode=barcode,
            measurement_time=data.get(str, "Measurement date"),
            measured_height=data.get(float, "Measured height"),
            chamber_temperature_at_start=data.get(
                float, "Chamber temperature at start"
            ),
            formula=formula,
            name=name.strip(),
        )


@dataclass(frozen=True)
class ResultPlateInfo(PlateInfo):
    label: str
    measinfo: str
    emission_filter_id: str

    @staticmethod
    def create(series: SeriesData) -> ResultPlateInfo | None:
        label = series.get(str, "Label")
        if label is None:
            return None

        measinfo = series.get(str, "Measinfo")
        if measinfo is None:
            return None

        plate_number = series[str, "Plate"]
        raw_barcode = series.get(str, "Barcode")
        barcode = (raw_barcode or '=""').removeprefix('="').removesuffix('"')
        barcode = barcode or f"Plate {plate_number}"

        return ResultPlateInfo(
            plate_number,
            barcode,
            series.get(str, "Measurement date"),
            series.get(float, "Measured height"),
            series.get(float, "Chamber temperature at start"),
            label=label,
            measinfo=measinfo,
            emission_filter_id=assert_not_none(
                search("De=(...)", measinfo),
                msg=f"Unable to find emission filter ID for Plate {barcode}.",
            ).group(1),
        )

    def match(self, background_info: BackgroundInfo) -> bool:
        return all(
            [
                background_info.plate_num == self.number,
                background_info.label in self.label,
                background_info.measinfo == self.measinfo,
            ]
        )


@dataclass
class BackgroundInfo:
    plate_num: str
    label: str
    measinfo: str


@dataclass
class BackgroundInfoList:
    background_info: list[BackgroundInfo]

    @staticmethod
    def create(reader: CsvReader) -> BackgroundInfoList | None:
        title = reader.pop_if_match("^Background information")
        if title is None:
            return None

        data_frame = assert_not_none(
            reader.pop_csv_block_as_df(header=0),
            "background information",
        )

        row_data = [SeriesData(series) for _, series in data_frame.iterrows()]

        return BackgroundInfoList(
            background_info=[
                BackgroundInfo(
                    plate_num=data[
                        str,
                        "Plate",
                        "Unable to find plate number from background info.",
                    ],
                    label=data[
                        str, "Label", "Unable to find label from background info."
                    ],
                    measinfo=data[
                        str,
                        "MeasInfo",
                        "Unable to find meas info from background info.",
                    ],
                )
                for data in row_data
            ]
        )


@dataclass
class CalculatedResult:
    uuid: str
    col: str
    row: str
    value: float


@dataclass
class CalculatedResultList:
    calculated_results: list[CalculatedResult]

    @staticmethod
    def create(reader: CsvReader) -> CalculatedResultList:
        # Calculated results may or may not have a title
        reader.pop_if_match("^Calculated results")

        data_frame = assert_not_none(
            reader.pop_csv_block_as_df(),
            "results data",
        )
        series = (
            data_frame.drop(0, axis=0).drop(0, axis=1)
            if data_frame.iloc[1, 0] == "A"
            else data_frame
        )
        rows, cols = series.shape
        series.index = [num_to_chars(i) for i in range(rows)]  # type: ignore[assignment]
        series.columns = [str(i).zfill(2) for i in range(1, cols + 1)]  # type: ignore[assignment]

        return CalculatedResultList(
            calculated_results=[
                CalculatedResult(
                    uuid=random_uuid_str(),
                    col=col,
                    row=row,
                    value=series.loc[col, row],
                )
                for col, row in series.stack().index
            ]
        )


@dataclass
class Result:
    uuid: str
    col: str
    row: str
    value: int


@dataclass
class ResultList:
    results: list[Result]

    @staticmethod
    def create(reader: CsvReader) -> ResultList:
        # Results may or may not have a title
        reader.pop_if_match("^Results")

        data_frame = assert_not_none(
            reader.pop_csv_block_as_df(),
            "reader data",
        )
        series = (
            data_frame.drop(0, axis=0).drop(0, axis=1)
            if data_frame.iloc[1, 0] == "A"
            else data_frame
        )
        rows, cols = series.shape
        series.index = [num_to_chars(i) for i in range(rows)]  # type: ignore[assignment]
        series.columns = [str(i).zfill(2) for i in range(1, cols + 1)]  # type: ignore[assignment]

        return ResultList(
            results=[
                Result(
                    uuid=random_uuid_str(),
                    col=col,
                    row=row,
                    value=int(series.loc[col, row]),
                )
                for col, row in series.stack().index
            ]
        )


@dataclass
class Plate:
    plate_info: CalculatedPlateInfo | ResultPlateInfo
    background_info_list: BackgroundInfoList | None
    calculated_result_list: CalculatedResultList
    result_list: ResultList

    @staticmethod
    def create(reader: CsvReader) -> Plate:
        series = PlateInfo.get_series(reader)
        if result_plate_info := ResultPlateInfo.create(series):
            return Plate(
                plate_info=result_plate_info,
                background_info_list=BackgroundInfoList.create(reader),
                calculated_result_list=CalculatedResultList([]),
                result_list=ResultList.create(reader),
            )
        else:
            return Plate(
                plate_info=CalculatedPlateInfo.create(series),
                background_info_list=BackgroundInfoList.create(reader),
                calculated_result_list=CalculatedResultList.create(reader),
                result_list=ResultList([]),
            )

    def collect_result_plates(self, plate_list: PlateList) -> list[Plate]:
        background_info_list = assert_not_none(
            self.background_info_list,
            msg=f"Unable to collect result plates, there is no background information for plate {self.plate_info.number}",
        )

        return [
            assert_not_none(
                plate_list.get_result_plate(background_info),
                msg=f"Unable to find result plate {background_info.label}.",
            )
            for background_info in background_info_list.background_info
        ]


@dataclass
class PlateList:
    plates: list[Plate]

    @staticmethod
    def create(reader: CsvReader) -> PlateList:
        plates: list[Plate] = []

        while reader.match("^Plate information"):
            plates.append(Plate.create(reader))
        return PlateList(plates)

    def get_result_plate(self, background_info: BackgroundInfo) -> Plate | None:
        for plate in self.plates:
            if isinstance(plate.plate_info, CalculatedPlateInfo):
                continue

            if plate.plate_info.match(background_info):
                return plate
        return None


@dataclass(frozen=True)
class BasicAssayInfo:
    protocol_id: str | None
    assay_id: str | None

    @staticmethod
    def create(reader: CsvReader) -> BasicAssayInfo:
        reader.drop_until_inclusive("^Basic assay information")
        data_frame = assert_not_none(
            reader.pop_csv_block_as_df(),
            "Basic assay information",
        ).T
        data_frame.iloc[0] = data_frame.iloc[0].replace(":.*", "", regex=True)
        data = SeriesData(df_to_series(data_frame))
        return BasicAssayInfo(data.get(str, "Protocol ID"), data.get(str, "Assay ID"))


@dataclass(frozen=True)
class PlateType:
    number_of_wells: float

    @staticmethod
    def create(reader: CsvReader) -> PlateType:
        reader.drop_until_inclusive("^Plate type")
        data_frame = assert_not_none(reader.pop_csv_block_as_df(), "Plate type").T
        data = SeriesData(df_to_series(data_frame))
        return PlateType(
            number_of_wells=data[float, "Number of the wells in the plate"]
        )


def get_sample_role_type(encoding: str) -> SampleRoleType:
    # BL        blank               blank_role
    # CTL       control             control_sample_role
    # LB        lance_blank         blank_role
    # LC        lance_crosstalk     control_sample_role
    # LH        lance_high          control_sample_role
    # S         pl_sample           sample_role
    # STD       standard            standard_sample_role
    # -         unknown             unknown_sample_role
    # UNK       unknown             unknown_sample_role
    # ZH        z_high              control_sample_role
    # ZL        z_low               control_sample_role
    sample_role_type_map = {
        "BL": SampleRoleType.blank_role,
        "CTL": SampleRoleType.control_sample_role,
        "LB": SampleRoleType.blank_role,
        "LC": SampleRoleType.control_sample_role,
        "LH": SampleRoleType.control_sample_role,
        "STD": SampleRoleType.standard_sample_role,
        "S": SampleRoleType.sample_role,
        "-": SampleRoleType.unknown_sample_role,
        "UNK": SampleRoleType.unknown_sample_role,
        "ZH": SampleRoleType.control_sample_role,
        "ZL": SampleRoleType.control_sample_role,
    }
    for pattern, value in sample_role_type_map.items():
        if encoding.startswith(pattern):
            return value

    msg = f"Unable to determine sample role type of plate map encoding {encoding}; expected to start with one of {sorted(sample_role_type_map.keys())}."
    raise AllotropeConversionError(msg)


@dataclass(frozen=True)
class PlateMap:
    plate_n: str
    group_n: str
    sample_role_type_mapping: dict[str, dict[str, SampleRoleType]]

    @staticmethod
    def create(reader: CsvReader) -> PlateMap | None:
        if not reader.current_line_exists() or reader.match("^Calculations"):
            return None

        *_, plate_n = assert_not_none(reader.pop(), "Platemap number").split(",")
        *_, group_n = assert_not_none(reader.pop(), "Platemap group").split(",")

        data_frame = assert_not_none(
            reader.pop_csv_block_as_df(),
            "Platemap data",
        ).replace(" ", "", regex=True)

        reader.pop_data()  # drop type specification
        reader.drop_empty()

        series = (
            data_frame.drop(0, axis=0).drop(0, axis=1)
            if data_frame.iloc[1, 0] == "A"
            else data_frame
        )
        rows, cols = series.shape
        series.index = [num_to_chars(i) for i in range(rows)]  # type: ignore[assignment]
        series.columns = [str(i).zfill(2) for i in range(1, cols + 1)]  # type: ignore[assignment]

        sample_role_type_mapping: dict[str, dict[str, SampleRoleType]] = {}
        for row, row_data in series.replace([np.nan, "''"], None).to_dict().items():
            col_mapping: dict[str, SampleRoleType] = {}
            for col, value in row_data.items():
                if value:
                    if role_type := get_sample_role_type(str(value)):
                        col_mapping[str(col)] = role_type
            if col_mapping:
                sample_role_type_mapping[str(row)] = col_mapping

        return PlateMap(plate_n, group_n, sample_role_type_mapping)

    def get_sample_role_type(self, col: str, row: str) -> SampleRoleType:
        try:
            return self.sample_role_type_mapping[row][col]
        except KeyError as e:
            msg = (
                f"Invalid plate map location for plate map {self.plate_n}: {col} {row}."
            )
            raise AllotropeConversionError(msg) from e


def create_plate_maps(reader: CsvReader) -> dict[str, PlateMap]:
    assert_not_none(
        reader.drop_until("^Platemap"),
        msg="No 'Platemap' section found.",
    )

    reader.pop()  # remove title

    maps: dict[str, PlateMap] = {}
    while _map := PlateMap.create(reader):
        maps[_map.plate_n] = _map
    return maps


@dataclass(frozen=True)
class Filter:
    name: str
    wavelength: float
    bandwidth: float | None = None

    @staticmethod
    def create(reader: CsvReader) -> Filter | None:
        if not reader.current_line_exists() or reader.match(
            "(^Mirror modules)|(^Instrument:)|(^Aperture:)"
        ):
            return None

        data_frame = assert_not_none(
            reader.pop_csv_block_as_df(),
            "Filter information",
        )
        data = SeriesData(df_to_series(data_frame.T))
        name = str(data.series.index[0])
        description = data[str, "Description"]

        if search_result := search("Longpass=(\\d+)nm", description):
            return Filter(name, wavelength=float(search_result.group(1)))
        return Filter(
            name,
            wavelength=float(
                assert_not_none(
                    search("CWL=(\\d+)nm", description),
                    msg=f"Unable to find wavelength for filter {name}.",
                ).group(1)
            ),
            bandwidth=float(
                assert_not_none(
                    search("BW=(\\d+)nm", description),
                    msg=f"Unable to find bandwidth for filter {name}.",
                ).group(1)
            ),
        )


def create_filters(reader: CsvReader) -> dict[str, Filter]:
    reader.drop_until("(^Filters:)|^Instrument:")

    if reader.match("^Instrument"):
        return {}

    reader.pop()  # remove title

    filters = {}
    while _filter := Filter.create(reader):
        filters[_filter.name] = _filter
    return filters


@dataclass(frozen=True)
class Labels:
    label: str
    excitation_filter: Filter | None
    emission_filters: dict[str, Filter | None]
    scan_position_setting: ScanPositionSettingPlateReader | None = None
    number_of_flashes: float | None = None
    detector_gain_setting: str | None = None

    @staticmethod
    def create(reader: CsvReader) -> Labels:
        reader.drop_until_inclusive("^Labels")
        data_frame = assert_not_none(
            reader.pop_csv_block_as_df(),
            "Labels",
        )
        data = SeriesData(df_to_series(data_frame.T).replace(np.nan, None))
        filters = create_filters(reader)
        filter_position_map = {
            "Bottom": ScanPositionSettingPlateReader.bottom_scan_position__plate_reader_,
            "Top": ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
        }

        return Labels(
            label=data.series.index[0],
            excitation_filter=filters.get(data.get(str, "Exc. filter", NOT_APPLICABLE)),
            emission_filters={
                "1st": filters.get(
                    data.get(str, "Ems. filter", NOT_APPLICABLE),
                ),
                "2nd": filters.get(data.get(str, "2nd ems. filter", NOT_APPLICABLE)),
            },
            scan_position_setting=filter_position_map.get(
                data.get(str, "Using of emission filter", NOT_APPLICABLE)
            ),
            number_of_flashes=data.get(float, "Number of flashes"),
            detector_gain_setting=data.get(str, "Reference AD gain"),
        )

    def get_emission_filter(self, id_val: str) -> Filter | None:
        return self.emission_filters.get(id_val)


@dataclass(frozen=True)
class Instrument:
    serial_number: str
    nickname: str

    @staticmethod
    def create(reader: CsvReader) -> Instrument:
        assert_not_none(
            reader.drop_until("^Instrument"),
            msg="No 'Instrument' section found.",
        )

        reader.pop()  # remove title

        *_, serial_number = assert_not_none(reader.pop(), "serial number").split(",")
        *_, nickname = assert_not_none(reader.pop(), "nickname").split(",")

        return Instrument(serial_number, nickname)


@dataclass(frozen=True)
class Software:
    software_name: str
    software_version: str

    @staticmethod
    def create(reader: CsvReader) -> Software:
        exported_with_regex = "Exported with (.+) version (.+)"
        assert_not_none(
            reader.drop_until(exported_with_regex),
            msg="Unable to find software information; no 'Exported with' section found.",
        )

        search_result = assert_not_none(
            search(
                exported_with_regex,
                assert_not_none(
                    reader.pop(),
                    "software information",
                ),
            )
        )

        return Software(
            software_name=search_result.group(1),
            software_version=search_result.group(2),
        )


@dataclass(frozen=True)
class Data:
    software: Software
    plate_list: PlateList
    basic_assay_info: BasicAssayInfo
    number_of_wells: float
    plate_maps: dict[str, PlateMap]
    labels: Labels
    instrument: Instrument

    @staticmethod
    def create(reader: CsvReader) -> Data:
        return Data(
            plate_list=PlateList.create(reader),
            basic_assay_info=BasicAssayInfo.create(reader),
            number_of_wells=PlateType.create(reader).number_of_wells,
            plate_maps=create_plate_maps(reader),
            labels=Labels.create(reader),
            instrument=Instrument.create(reader),
            software=Software.create(reader),
        )



from collections.abc import Iterable

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import ContainerType
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.pcr.BENCHLING._2023._09.qpcr import (
    CalculatedData,
    CalculatedDataItem,
    Data,
    DataCube,
    DataCubeComponent,
    DataSource,
    Measurement,
    MeasurementGroup,
    Metadata,
    ProcessedData,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_calculated_documents import (
    iter_calculated_data_documents,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_structure import (
    AmplificationData,
    Header,
    MeltCurveRawData,
    MulticomponentData,
    RawData,
    Result,
    ResultMetadata,
    Well,
    WellItem,
)
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.utils.calculated_data_documents.definition import (
    CalculatedDocument,
)
from allotropy.parsers.utils.values import try_int_or_nan


def _create_processed_data(
    amplification_data: AmplificationData,
    result: Result,
) -> ProcessedData:
    cycle_count = DataCubeComponent(FieldComponentDatatype.integer, "cycle count", "#")
    return ProcessedData(
        automatic_cycle_threshold_enabled_setting=result.automatic_baseline_determination_enabled_setting,
        cycle_threshold_value_setting=result.cycle_threshold_value_setting,
        automatic_baseline_determination_enabled_setting=result.automatic_baseline_determination_enabled_setting,
        genotyping_determination_method_setting=result.genotyping_determination_method_setting,
        genotyping_determination_result=result.genotyping_determination_result,
        cycle_threshold_result=result.cycle_threshold_result,
        normalized_reporter_result=result.normalized_reporter_result,
        baseline_corrected_reporter_result=result.baseline_corrected_reporter_result,
        data_cubes=[
            DataCube(
                label="normalized reporter",
                structure_dimensions=[cycle_count],
                structure_measures=[
                    DataCubeComponent(
                        FieldComponentDatatype.double,
                        "normalized report result",
                        UNITLESS,
                    )
                ],
                dimensions=[amplification_data.cycle],
                measures=[amplification_data.rn],
            ),
            DataCube(
                label="baseline corrected reporter",
                structure_dimensions=[cycle_count],
                structure_measures=[
                    DataCubeComponent(
                        FieldComponentDatatype.double,
                        "baseline corrected reporter result",
                        UNITLESS,
                    )
                ],
                dimensions=[amplification_data.cycle],
                measures=[amplification_data.delta_rn],
            ),
        ],
    )


def _create_measurement(
    well_item: WellItem,
    header: Header,
    multicomponent_data: MulticomponentData | None,
    melt_curve_raw_data: MeltCurveRawData | None,
    amplification_data: AmplificationData,
    result: Result,
) -> Measurement:
    # TODO: temp workaround for cal doc result
    well_item._result = result

    cycle_count = DataCubeComponent(FieldComponentDatatype.integer, "cycle count", "#")
    data_cubes = []
    if multicomponent_data:
        if well_item.reporter_dye_setting is not None:
            data_cubes.append(
                DataCube(
                    label="reporter dye",
                    structure_dimensions=[cycle_count],
                    structure_measures=[
                        DataCubeComponent(
                            FieldComponentDatatype.double,
                            "reporter dye fluorescence",
                            "RFU",
                        )
                    ],
                    dimensions=[multicomponent_data.cycle],
                    measures=[
                        multicomponent_data.get_column(well_item.reporter_dye_setting)
                    ],
                ),
            )
        if header.passive_reference_dye_setting is not None:
            data_cubes.append(
                DataCube(
                    label="passive reference dye",
                    structure_dimensions=[cycle_count],
                    structure_measures=[
                        DataCubeComponent(
                            FieldComponentDatatype.double,
                            "passive reference dye fluorescence",
                            "RFU",
                        )
                    ],
                    dimensions=[multicomponent_data.cycle],
                    measures=[
                        multicomponent_data.get_column(
                            header.passive_reference_dye_setting
                        )
                    ],
                ),
            )

    if melt_curve_raw_data:
        data_cubes.append(
            DataCube(
                label="melting curve",
                structure_dimensions=[
                    DataCubeComponent(
                        FieldComponentDatatype.double, "temperature", "degrees C"
                    )
                ],
                structure_measures=[
                    DataCubeComponent(
                        FieldComponentDatatype.double,
                        "reporter dye fluorescence",
                        UNITLESS,
                    ),
                    DataCubeComponent(FieldComponentDatatype.double, "slope", UNITLESS),
                ],
                dimensions=[melt_curve_raw_data.reading],
                measures=[
                    melt_curve_raw_data.fluorescence,
                    melt_curve_raw_data.derivative,
                ],
            ),
        )

    return Measurement(
        identifier=well_item.uuid,
        timestamp=header.measurement_time,
        target_identifier=well_item.target_dna_description,
        sample_identifier=well_item.sample_identifier,
        sample_role_type=well_item.sample_role_type,
        well_location_identifier=well_item.well_location_identifier,
        well_plate_identifier=header.barcode,
        total_cycle_number_setting=amplification_data.total_cycle_number_setting,
        pcr_detection_chemistry=header.pcr_detection_chemistry,
        reporter_dye_setting=well_item.reporter_dye_setting,
        quencher_dye_setting=well_item.quencher_dye_setting,
        passive_reference_dye_setting=header.passive_reference_dye_setting,
        processed_data=_create_processed_data(amplification_data, result),
        data_cubes=data_cubes,
    )


def _create_metadata(header: Header, file_name: str) -> Metadata:
    return Metadata(
        device_identifier=header.device_identifier,
        device_type="qPCR",
        device_serial_number=header.device_serial_number,
        model_number=header.model_number,
        software_name="Thermo QuantStudio",
        software_version="1.0",
        data_system_instance_identifier="localhost",
        file_name=file_name,
        unc_path="",  # unknown
        measurement_method_identifier=header.measurement_method_identifier,
        experiment_type=header.experiment_type,
        container_type=ContainerType.qPCR_reaction_block,
    )


def _create_calculated_data(
    calculated_data_documents: Iterable[CalculatedDocument],
    results_metadata: ResultMetadata,
) -> CalculatedData:
    return CalculatedData(
        reference_sample_description=results_metadata.reference_sample_description,
        reference_dna_description=results_metadata.reference_dna_description,
        items=[
            CalculatedDataItem(
                identifier=cal_doc.uuid,
                name=cal_doc.name,
                value=cal_doc.value,
                unit=UNITLESS,
                data_sources=[
                    DataSource(
                        identifier=data_source.reference.uuid,
                        feature=data_source.feature,
                    )
                    for data_source in cal_doc.data_sources
                ],
            )
            for cal_doc in calculated_data_documents
        ],
    )


def _create_data(
    file_name: str,
    header: Header,
    wells: list[Well],
    amp_data: dict[int, dict[str, AmplificationData]],
    multi_data: dict[int, MulticomponentData],
    results_data: dict[int, dict[str, Result]],
    results_metadata: ResultMetadata,
    melt_data: dict[int, MeltCurveRawData],
    calculated_documents: Iterable[CalculatedDocument],
) -> Data:
    measurement_groups = [
        MeasurementGroup(
            analyst=header.analyst,
            experimental_data_identifier=header.experimental_data_identifier,
            plate_well_count=try_int_or_nan(header.plate_well_count),
            measurements=[
                _create_measurement(
                    well_item,
                    header,
                    multi_data.get(well.identifier),
                    melt_data.get(well.identifier),
                    amp_data[well_item.identifier][well_item.target_dna_description],
                    results_data[well_item.identifier][
                        well_item.target_dna_description.replace(" ", "")
                    ],
                )
                for well_item in well.items
            ],
        )
        for well in wells
    ]

    return Data(
        metadata=_create_metadata(header, file_name),
        measurement_groups=measurement_groups,
        calculated_data=_create_calculated_data(calculated_documents, results_metadata),
    )


def create_data(reader: LinesReader, file_name: str) -> Data:
    # Data sections must be read in order from the file.
    header = Header.create(reader)
    wells = Well.create(reader, header.experiment_type)
    # Skip raw data section
    RawData.create(reader)
    amp_data = AmplificationData.create(reader)
    multi_data = MulticomponentData.create(reader)
    results_data, results_metadata = Result.create(reader, header.experiment_type)
    melt_data = MeltCurveRawData.create(reader)

    calculated_data_documents = iter_calculated_data_documents(
        [well_item for well in wells for well_item in well.items],
        header.experiment_type,
        results_metadata.reference_sample_description,
        results_metadata.reference_dna_description,
    )

    return _create_data(
        file_name,
        header,
        wells,
        amp_data,
        multi_data,
        results_data,
        results_metadata,
        melt_data,
        calculated_data_documents,
    )



import json
from pathlib import Path
from typing import Any

from allotropy.allotrope.schema_parser.path_util import (
    get_full_schema_path,
    get_schema_path_from_manifest,
    SHARED_SCHEMAS_DEFINITIONS_PATH,
)
from allotropy.constants import DEFAULT_ENCODING


def get_shared_definitions() -> dict[str, Any]:
    with open(Path(SHARED_SCHEMAS_DEFINITIONS_PATH, "definitions.json")) as f:
        return json.load(f)  # type: ignore[no-any-return]


def get_shared_unit_definitions() -> dict[str, Any]:
    with open(Path(SHARED_SCHEMAS_DEFINITIONS_PATH, "units.json")) as f:
        return json.load(f)  # type: ignore[no-any-return]


def add_definitions(schema: dict[str, Any]) -> dict[str, Any]:
    for file, section in [
        ("definitions", "defs"),
        ("units", "defs"),
        ("custom", "custom"),
    ]:
        existing = schema.get(f"${section}", {})
        with open(
            Path(SHARED_SCHEMAS_DEFINITIONS_PATH, f"{file}.json"),
            encoding=DEFAULT_ENCODING,
        ) as f:
            additional = json.load(f)
        additional.update(existing)
        schema[f"${section}"] = additional
    return schema


def get_schema(schema_path: Path) -> dict[str, Any]:
    with open(get_full_schema_path(schema_path)) as f:
        return add_definitions(json.load(f))


def get_schema_from_manifest(manifest: str) -> dict[str, Any]:
    return get_schema(get_schema_path_from_manifest(manifest))


def get_schema_from_model(model: Any) -> dict[str, Any]:
    manifest = getattr(model, "manifest", getattr(model, "field_asm_manifest", None))
    if not manifest:
        msg = f"No 'manifest' or 'field_asm_manifest' found in model: {type(model)}"
        raise ValueError(msg)
    return get_schema_from_manifest(manifest)



from typing import Any

from allotropy.allotrope.models.shared.definitions.custom import (
    TNullableQuantityValueCell,
    TNullableQuantityValueGramPerLiter,
    TNullableQuantityValueMicrometer,
    TNullableQuantityValueMillimeterOfMercury,
    TNullableQuantityValueMillimolePerLiter,
    TNullableQuantityValueMillionCellsPerMilliliter,
    TNullableQuantityValueMilliOsmolesPerKilogram,
    TNullableQuantityValueOpticalDensity,
    TNullableQuantityValuePercent,
    TNullableQuantityValuePH,
    TNullableQuantityValueTODO,
    TNullableQuantityValueUnitPerLiter,
)

FILENAME_REGEX = r"^SampleResults(?P<device_identifier>.*)?(?P<processing_time>\d{4}-\d{2}-\d{2}_\d{6})\.csv$"

INVALID_FILENAME_MESSAGE = (
    "{} is not valid. File name is expected to have format of SampleResultsYYYY-MM-DD_HHMMSS.csv "
    "or SampleResults<Analyzer ID>YYYY-MM-DD_HHMMSS.csv where <Analyzer ID> is defined in Settings"
)

MOLAR_CONCENTRATION_CLASSES = (
    TNullableQuantityValueMillimolePerLiter
    | TNullableQuantityValueGramPerLiter
    | TNullableQuantityValueUnitPerLiter
)
ALL_MOLAR_CONCENTRATION_CLASSES: list[
    (
        type[TNullableQuantityValueMillimolePerLiter]
        | type[TNullableQuantityValueGramPerLiter]
        | type[TNullableQuantityValueUnitPerLiter]
    )
] = [
    TNullableQuantityValueMillimolePerLiter,
    TNullableQuantityValueGramPerLiter,
    TNullableQuantityValueUnitPerLiter,
]
MOLAR_CONCENTRATION_CLS_BY_UNIT = {
    cls.unit: cls for cls in ALL_MOLAR_CONCENTRATION_CLASSES
}

ANALYTE_MAPPINGS: dict[str, dict[str, str]] = {
    "NH4+": {
        "name": "ammonium",
        "unit": "mmol/L",
    },
    "HCO3": {
        "name": "bicarbonate",
        "unit": "mmol/L",
    },
    "Ca++": {
        "name": "calcium",
        "unit": "mmol/L",
    },
    "Gluc": {
        "name": "glucose",
        "unit": "g/L",
    },
    "Glu": {
        "name": "glutamate",
        "unit": "mmol/L",
    },
    "Gln": {
        "name": "glutamine",
        "unit": "mmol/L",
    },
    "Lac": {
        "name": "lactate",
        "unit": "g/L",
    },
    "K+": {
        "name": "potassium",
        "unit": "mmol/L",
    },
    "Na+": {
        "name": "sodium",
        "unit": "mmol/L",
    },
}

PROPERTY_MAPPINGS: dict[str, dict[str, Any]] = {
    "pco2": {
        "col_name": "PCO2",
        "cls": TNullableQuantityValueMillimeterOfMercury,
    },
    "co2_saturation": {
        "col_name": "CO2 Saturation",
        "cls": TNullableQuantityValuePercent,
    },
    "po2": {
        "col_name": "PO2",
        "cls": TNullableQuantityValueMillimeterOfMercury,
    },
    "o2_saturation": {
        "col_name": "O2 Saturation",
        "cls": TNullableQuantityValuePercent,
    },
    "optical_density": {
        "col_name": "Optical Density",
        "cls": TNullableQuantityValueOpticalDensity,
    },
    "pH": {
        "col_name": "pH",
        "cls": TNullableQuantityValuePH,
    },
    "osmolality": {
        "col_name": "Osm",
        "cls": TNullableQuantityValueMilliOsmolesPerKilogram,
    },
    "viability__cell_counter_": {
        "col_name": "Viability",
        "cls": TNullableQuantityValuePercent,
    },
    "total_cell_density__cell_counter_": {
        "col_name": "Total Density",
        "cls": TNullableQuantityValueMillionCellsPerMilliliter,
    },
    "viable_cell_density__cell_counter_": {
        "col_name": "Viable Density",
        "cls": TNullableQuantityValueMillionCellsPerMilliliter,
    },
    "average_live_cell_diameter__cell_counter_": {
        "col_name": "Average Live Cell Diameter",
        "cls": TNullableQuantityValueMicrometer,
    },
    "average_total_cell_diameter__cell_counter_": {
        "col_name": "Average Total Cell Diameter",
        "cls": TNullableQuantityValueMicrometer,
    },
    "total_cell_diameter_distribution__cell_counter_": {
        "col_name": "Total Cell Diameter Distribution",
        "cls": TNullableQuantityValueTODO,
    },
    "viable_cell_count__cell_counter_": {
        "col_name": "Total Live Count",
        "cls": TNullableQuantityValueCell,
    },
    "total_cell_count__cell_counter_": {
        "col_name": "Total Cell Count",
        "cls": TNullableQuantityValueCell,
    },
}



from __future__ import annotations

from dataclasses import dataclass

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import (
    JsonFloat,
    NaN,
)
from allotropy.parsers.chemometec_nucleoview.constants import (
    DEFAULT_ANALYST,
    DEFAULT_EPOCH_TIMESTAMP,
    DEFAULT_MODEL_NUMBER,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData


@dataclass
class Row:
    model_number: str
    equipment_serial_number: str | None
    software_version: str | None
    analyst: str
    timestamp: str
    sample_identifier: str
    multiplication_factor: float | None
    viability_percent: float
    live_cell_count: JsonFloat
    dead_cell_count: JsonFloat
    total_cell_count: JsonFloat
    estimated_cell_diameter: JsonFloat

    @staticmethod
    def create(data: SeriesData) -> Row | None:
        if data.get(float, "Viability (%)", validate=SeriesData.NOT_NAN) is None:
            return None

        timestamp = data.get(str, "Date time")
        if timestamp:
            offset = data.get(str, "Time zone offset", "0")
            timestamp = pd.to_datetime(
                f"{timestamp}{'+' if offset[0] not in {'+', '-'} else ''}{offset}"
            ).isoformat()

        return Row(
            model_number=data.get(str, "Instrument type", DEFAULT_MODEL_NUMBER),
            equipment_serial_number=data.get(str, "Instrument s/n"),
            software_version=data.get(str, "Application SW version"),
            analyst=data.get(str, "Operator", DEFAULT_ANALYST),
            timestamp=timestamp or DEFAULT_EPOCH_TIMESTAMP,
            sample_identifier=data[str, "Image"].split("-")[3],
            multiplication_factor=data.get(float, "Multiplication factor"),
            viability_percent=data[float, "Viability (%)"],
            # Cell counts are measured in cells/mL, but reported in millions of cells/mL
            live_cell_count=data.get(float, "Live (cells/ml)", NaN) / 1e6,
            dead_cell_count=data.get(float, "Dead (cells/ml)", NaN) / 1e6,
            total_cell_count=data.get(float, "Total (cells/ml)", NaN) / 1e6,
            estimated_cell_diameter=data.get(
                float, "Estimated cell diameter (um)", NaN
            ),
        )

    @staticmethod
    def create_rows(data: pd.DataFrame) -> list[Row]:
        return [row for row in map_rows(data, Row.create) if row]



from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.thermo_fisher_nanodrop_eight.nanodrop_eight_reader import (
    NanoDropEightReader,
)
from allotropy.parsers.thermo_fisher_nanodrop_eight.nanodrop_eight_structure import (
    create_data,
)
from allotropy.parsers.vendor_parser import VendorParser


class NanodropEightParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Thermo Fisher NanoDrop Eight"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        df = NanoDropEightReader.read(named_file_contents)
        data = create_data(df, named_file_contents.original_file_name)
        return self._get_mapper(Mapper).map_model(data)



from __future__ import annotations

from typing import Any
import xml.etree.ElementTree as Et

from allotropy.allotrope.models.adm.multi_analyte_profiling.benchling._2024._01.multi_analyte_profiling import (
    AnalyteAggregateDocument,
    AnalyteDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    ErrorAggregateDocument,
    ErrorDocumentItem,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    MultiAnalyteProfilingAggregateDocument,
    MultiAnalyteProfilingDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMicroliter,
    TQuantityValueNumber,
    TQuantityValueRelativeFluorescenceUnit,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TStatisticDatumRole,
)
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.exceptions import (
    get_key_or_error,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.biorad_bioplex_manager.biorad_bioplex_manager_structure import (
    AnalyteDocumentData,
    DeviceWellSettings,
    get_well_name,
    SampleDocumentAggregate,
    SampleDocumentStructure,
    validate_xml_structure,
    WellAnalyteMapping,
    WellSystemLevelMetadata,
)
from allotropy.parsers.biorad_bioplex_manager.constants import (
    BEAD_REGIONS,
    CONTAINER_TYPE,
    DESCRIPTION_TAG,
    DEVICE_TYPE,
    DOC_LOCATION_TAG,
    ERROR_MAPPING,
    PLATE_DIMENSIONS_TAG,
    PRODUCT_MANUFACTURER,
    SAMPLES,
    SOFTWARE_NAME,
    TOTAL_WELLS_ATTRIB,
    USER,
    VERSION_ATTRIB,
    WELLS_TAG,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import quantity_or_none
from allotropy.parsers.utils.xml import get_val_from_xml
from allotropy.parsers.vendor_parser import VendorParser


class BioradBioplexParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Bio-Rad Bio-Plex Manager"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        contents = named_file_contents.contents.read()
        xml_tree = Et.ElementTree(Et.fromstring(contents))  # noqa: S314
        root_xml = xml_tree.getroot()
        validate_xml_structure(root_xml)
        filename = named_file_contents.original_file_name
        return self._get_model(root_xml, filename)

    def _get_model(self, data: Et.Element, filename: str) -> Model:
        software_version_value = data.attrib[VERSION_ATTRIB]
        for child in data:
            if child.tag == SAMPLES:
                all_samples_xml = child
            elif child.tag == PLATE_DIMENSIONS_TAG:
                plate_well_count = int(child.attrib[TOTAL_WELLS_ATTRIB])
            elif child.tag == DOC_LOCATION_TAG:
                experimental_data_id = child.text
            elif child.tag == DESCRIPTION_TAG:
                experiment_type = child.text
        for child in data:
            if child.tag == WELLS_TAG:
                well_system_metadata = WellSystemLevelMetadata.create(child[0])
                device_document = BioradBioplexParser._get_device_system_document(
                    well_system_metadata
                )
                all_wells_xml = child
        multi_docs = self.get_measurement_document_aggregate(
            samples_xml=all_samples_xml,
            wells_xml=all_wells_xml,
            regions_of_interest=well_system_metadata.regions_of_interest,
            experimental_data_id=experimental_data_id,
            experiment_type=experiment_type,
            plate_well_count=plate_well_count,
            analytical_method_identifier=well_system_metadata.analytical_method,
            plate_id=well_system_metadata.plate_id,
        )

        return Model(
            field_asm_manifest="http://purl.allotrope.org/manifests/multi-analyte-profiling/BENCHLING/2024/01/multi-analyte-profiling.manifest",
            multi_analyte_profiling_aggregate_document=MultiAnalyteProfilingAggregateDocument(
                device_system_document=device_document,
                data_system_document=DataSystemDocument(
                    software_name=SOFTWARE_NAME,
                    software_version=software_version_value,
                    ASM_converter_name=self.get_asm_converter_name(),
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                    file_name=filename,
                ),
                multi_analyte_profiling_document=multi_docs,
            ),
        )

    @staticmethod
    def _get_device_system_document(
        well_system_metadata: WellSystemLevelMetadata,
    ) -> DeviceSystemDocument:
        return DeviceSystemDocument(
            equipment_serial_number=well_system_metadata.serial_number,
            firmware_version=well_system_metadata.controller_version,
            product_manufacturer=PRODUCT_MANUFACTURER,
        )

    def get_measurement_document_aggregate(
        self,
        samples_xml: Et.Element,
        wells_xml: Et.Element,
        regions_of_interest: list[str],
        experimental_data_id: str | None,
        experiment_type: str | None,
        plate_well_count: int,
        analytical_method_identifier: str,
        plate_id: str,
    ) -> list[MultiAnalyteProfilingDocumentItem]:
        sample_document_aggregated = SampleDocumentAggregate.create(samples_xml)
        multi_analyte_docs = []
        for well in wells_xml:
            well_name = get_well_name(well.attrib)
            sample = sample_document_aggregated.samples_dict[well_name]
            device_well_settings = DeviceWellSettings.create(well)
            measurement_doc = MeasurementDocumentItem(
                measurement_identifier=random_uuid_str(),
                measurement_time=self._get_date_time(
                    device_well_settings.acquisition_time
                ),
                assay_bead_count=TQuantityValueNumber(
                    value=device_well_settings.well_total_events
                ),
                sample_document=BioradBioplexParser._get_sample_document(
                    sample, well_name, plate_id
                ),
                device_control_aggregate_document=BioradBioplexParser._get_device_control_aggregate(
                    device_well_settings, sample
                ),
                analyte_aggregate_document=BioradBioplexParser._get_analyte_aggregate_document(
                    well,
                    sample_document_aggregated.analyte_region_dict,
                    regions_of_interest,
                ),
                error_aggregate_document=BioradBioplexParser._get_error_aggregate_document(
                    sample.well_analyte_mapping
                ),
            )
            multi_analyte_docs.append(
                MultiAnalyteProfilingDocumentItem(
                    measurement_aggregate_document=MeasurementAggregateDocument(
                        measurement_document=[measurement_doc],
                        experiment_type=experiment_type,
                        plate_well_count=TQuantityValueNumber(value=plate_well_count),
                        analytical_method_identifier=analytical_method_identifier,
                        container_type=CONTAINER_TYPE,
                        experimental_data_identifier=experimental_data_id,
                    ),
                    analyst=get_val_from_xml(well, USER),
                )
            )

        return multi_analyte_docs

    @staticmethod
    def _get_sample_document(
        sample: SampleDocumentStructure, well_name: str, well_plate_id: str
    ) -> Any:
        return SampleDocument(
            description=sample.description,
            sample_identifier=sample.sample_identifier,
            location_identifier=well_name,
            sample_role_type=sample.sample_type,
            well_plate_identifier=well_plate_id,
        )

    @staticmethod
    def _get_device_control_aggregate(
        device_well_settings: DeviceWellSettings, sample: SampleDocumentStructure
    ) -> DeviceControlAggregateDocument:

        device_control_doc_item = DeviceControlDocumentItem(
            device_type=DEVICE_TYPE,
            sample_volume_setting=TQuantityValueMicroliter(
                value=device_well_settings.sample_volume_setting
            ),
            dilution_factor_setting=quantity_or_none(
                TQuantityValueUnitless, sample.sample_dilution
            ),
            detector_gain_setting=device_well_settings.detector_gain_setting,
            minimum_assay_bead_count_setting=quantity_or_none(
                TQuantityValueNumber,
                device_well_settings.minimum_assay_bead_count_setting,
            ),
        )
        return DeviceControlAggregateDocument(
            device_control_document=[device_control_doc_item]
        )

    @staticmethod
    def _get_analyte_aggregate_document(
        well: Et.Element,
        analyte_region_dict: dict[str, str],
        regions_of_interest: list[str],
    ) -> AnalyteAggregateDocument:
        analyte_docs = []
        for atr in well:
            if atr.tag == BEAD_REGIONS:
                for bead in atr:
                    analyte_structure_doc = AnalyteDocumentData.create(
                        bead_region_xml=bead,
                        analyte_region_dict=analyte_region_dict,
                        regions_of_interest=regions_of_interest,
                    )
                    if analyte_structure_doc is None:
                        continue
                    analyte_docs.append(
                        AnalyteDocumentItem(
                            analyte_identifier=random_uuid_str(),
                            analyte_name=analyte_structure_doc.analyte_name,
                            assay_bead_identifier=analyte_structure_doc.assay_bead_identifier,
                            assay_bead_count=TQuantityValueNumber(
                                value=analyte_structure_doc.assay_bead_count
                            ),
                            fluorescence=TQuantityValueRelativeFluorescenceUnit(
                                value=analyte_structure_doc.fluorescence,
                                has_statistic_datum_role=TStatisticDatumRole.median_role,
                            ),
                        )
                    )
        return AnalyteAggregateDocument(analyte_document=analyte_docs)

    @staticmethod
    def _get_error_aggregate_document(
        well_analyte_mapping: WellAnalyteMapping,
    ) -> ErrorAggregateDocument:
        error_docs = []
        for analyte in well_analyte_mapping.analytes:
            if analyte.analyte_error_code != "0":
                error_docs.append(
                    ErrorDocumentItem(
                        error=analyte.analyte_name,
                        error_feature=get_key_or_error(
                            "error code", analyte.analyte_error_code, ERROR_MAPPING
                        ),
                    ),
                )
        return ErrorAggregateDocument(error_document=error_docs)



import numpy as np
import pandas as pd

from allotropy.types import IOType


class QiacuitydPCRReader:
    def __init__(self, contents: IOType):
        # Read in the file, skip first row since it does not have data in it.
        qiacuity_dpcr_data = pd.read_csv(filepath_or_buffer=contents, header=1).replace(
            np.nan, None
        )
        column_names = qiacuity_dpcr_data.columns.tolist()
        # Rename the blank column to specify that it's the Well Name column
        column_names[0] = "Well Name"
        column_index = pd.Index(column_names)
        qiacuity_dpcr_data.columns = column_index
        self.well_data = qiacuity_dpcr_data



from dataclasses import dataclass

from allotropy.types import IOType


@dataclass(frozen=True)
class NamedFileContents:
    """
    A file's contents, the name of the file whence they came, and an optional encoding.

    In order to support use cases where the file contents of a vendor file are not stored locally,
    we accept an IO object for file contents. However, some vendor file parsers extract information
    from the original file name produced by the instrument. Thus, we use this class to support
    as many cases as possible.

    An encoding of None means that an encoding was not specified. Currently, if encoding is not specified
    (and contents is a bytes object), we assume DEFAULT_ENCODING. If CHARDET_ENCODING is specified, we
    will [try to] use chardet to detect the encoding.
    """

    contents: IOType
    original_file_name: str
    encoding: str | None = None



from __future__ import annotations

from dataclasses import dataclass, field
from xml.etree import ElementTree

from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.exceptions import AllotropeConversionError, AllotropyParserError
from allotropy.parsers.biorad_bioplex_manager.constants import (
    ACQ_TIME,
    ANALYTE_NAME,
    BEAD_COUNT,
    CODE,
    COLUMN_NUMBER,
    DESCRIPTION_TAG,
    DILUTION,
    DOC_LOCATION_TAG,
    LABEL,
    MACHINE_INFO,
    MEDIAN,
    MEMBER_WELLS,
    MW_ANALYTES,
    PLATE_DIMENSIONS_TAG,
    READING,
    REGION_COUNT,
    REGION_NUMBER,
    REGIONS_OF_INTEREST,
    ROW_NAMES,
    ROW_NUMBER,
    RP1_GAIN,
    RUN_CONDITIONS,
    RUN_SETTINGS,
    SAMPLE_VOLUME,
    SAMPLES,
    SERIAL_NUMBER,
    STOP_READING_CRITERIA,
    TOTAL_EVENTS,
    WELLS_TAG,
)
from allotropy.parsers.utils.values import (
    try_float,
    try_float_or_none,
    try_int,
)
from allotropy.parsers.utils.xml import (
    get_attrib_from_xml,
    get_element_from_xml,
    get_val_from_xml,
    get_val_from_xml_or_none,
)

SAMPLE_ROLE_TYPE_MAPPING = {
    "Blank": SampleRoleType.blank_role,
    "Control": SampleRoleType.control_sample_role,
    "Standard": SampleRoleType.standard_sample_role,
    "Unknown": SampleRoleType.unknown_sample_role,
}


@dataclass(frozen=True)
class AnalyteSample:
    analyte_name: str
    analyte_region: int
    analyte_error_code: str

    @staticmethod
    def create(analyte_xml: ElementTree.Element) -> AnalyteSample:
        return AnalyteSample(
            analyte_name=get_val_from_xml(analyte_xml, ANALYTE_NAME),
            analyte_region=try_int(analyte_xml.attrib[REGION_NUMBER], "analyte_region"),
            analyte_error_code=get_attrib_from_xml(analyte_xml, READING, CODE),
        )


@dataclass(frozen=True)
class WellAnalyteMapping:
    well_name: str
    analytes: list[AnalyteSample] = field(default_factory=list)


@dataclass
class SampleDocumentStructure:
    sample_type: SampleRoleType
    sample_identifier: str
    description: str | None
    well_name: str
    sample_dilution: float | None
    well_analyte_mapping: WellAnalyteMapping


@dataclass
class SampleDocumentAggregate:
    # This data class pulled from the <Samples> part of the xml.
    samples_dict: dict[str, SampleDocumentStructure] = field(default_factory=dict)
    # Default to empty dictionary.
    analyte_region_dict: dict[str, str] = field(default_factory=dict)

    @staticmethod
    def create(samples_xml: ElementTree.Element) -> SampleDocumentAggregate:
        sample_documents = SampleDocumentAggregate()
        for sample_types in samples_xml:
            for child_sample_type in sample_types:
                sample_type = map_sample_type(child_sample_type.tag)
                sample_description = get_val_from_xml_or_none(
                    child_sample_type, DESCRIPTION_TAG
                )
                sample_identifier = get_val_from_xml(child_sample_type, LABEL)
                sample_dilution = get_val_from_xml_or_none(child_sample_type, DILUTION)
                for child in child_sample_type:
                    if child.tag == MEMBER_WELLS:
                        for member_well in child:
                            sample_document = (
                                SampleDocumentAggregate._generate_sample_document(
                                    sample_documents=sample_documents,
                                    well_xml=member_well,
                                    sample_id=sample_identifier,
                                    sample_dilution=try_float_or_none(sample_dilution),
                                    sample_type=sample_type,
                                    sample_description=sample_description,
                                )
                            )
                            sample_documents.samples_dict[
                                sample_document.well_name
                            ] = sample_document
        return sample_documents

    @staticmethod
    def _generate_sample_document(
        sample_documents: SampleDocumentAggregate,
        well_xml: ElementTree.Element,
        sample_id: str,
        sample_dilution: float | None,
        sample_type: SampleRoleType,
        sample_description: str | None,
    ) -> SampleDocumentStructure:
        well_name = get_well_name(well_xml.attrib)
        mappings = WellAnalyteMapping(well_name=well_name, analytes=[])
        for analyte in get_element_from_xml(well_xml, MW_ANALYTES):
            # Create the analyte.
            new_analyte = AnalyteSample.create(analyte_xml=analyte)
            # Add the analyte to the well analyte mappings.
            mappings.analytes.append(new_analyte)
            # Update the analyte region dict
            sample_documents.analyte_region_dict[
                str(new_analyte.analyte_region)
            ] = new_analyte.analyte_name
        sample_document = SampleDocumentStructure(
            sample_type=sample_type,
            sample_identifier=sample_id,
            description=sample_description,
            well_name=well_name,
            well_analyte_mapping=mappings,
            sample_dilution=sample_dilution,
        )
        return sample_document


@dataclass
class DeviceWellSettings:
    # This data class is for all metadata needed in the <Wells> section, used in measurement and device control docs.
    well_name: str
    sample_volume_setting: int
    detector_gain_setting: str
    minimum_assay_bead_count_setting: int
    well_total_events: int
    acquisition_time: str

    @staticmethod
    def create(well_xml: ElementTree.Element) -> DeviceWellSettings:
        well_name = get_well_name(well_xml.attrib)
        well_acq_time = get_val_from_xml(well_xml, ACQ_TIME)
        total_events = get_val_from_xml(well_xml, TOTAL_EVENTS)
        sample_volume = try_int(
            get_val_from_xml(well_xml, RUN_SETTINGS, SAMPLE_VOLUME), "sample_volume"
        )

        detector_gain_setting = get_val_from_xml(well_xml, RUN_CONDITIONS, RP1_GAIN)
        min_assay_bead_count_setting = get_attrib_from_xml(
            well_xml, RUN_SETTINGS, BEAD_COUNT, STOP_READING_CRITERIA
        )
        return DeviceWellSettings(
            well_name=well_name,
            sample_volume_setting=sample_volume,
            detector_gain_setting=detector_gain_setting,
            minimum_assay_bead_count_setting=try_int(
                min_assay_bead_count_setting, "minimum_assay_bead_count_settings"
            ),
            acquisition_time=well_acq_time,
            well_total_events=try_int(total_events, "well_total_events"),
        )


@dataclass
class AnalyteDocumentData:
    analyte_name: str
    assay_bead_identifier: str
    assay_bead_count: int
    fluorescence: float

    @staticmethod
    def create(
        bead_region_xml: ElementTree.Element,
        analyte_region_dict: dict[str, str],
        regions_of_interest: list[str],
    ) -> AnalyteDocumentData | None:
        # Look up analyte name from sample
        assay_bead_identifier = bead_region_xml.attrib[REGION_NUMBER]

        # Look up bead region -> analyte name
        if assay_bead_identifier in regions_of_interest:
            analyte_name = analyte_region_dict[assay_bead_identifier]
            assay_bead_count = try_int(
                get_val_from_xml(bead_region_xml, REGION_COUNT), "assay_bead_count"
            )
            # Median
            fluorescence = try_float(
                get_val_from_xml(bead_region_xml, MEDIAN), "fluorescence"
            )
            return AnalyteDocumentData(
                analyte_name=analyte_name,
                assay_bead_identifier=assay_bead_identifier,
                assay_bead_count=assay_bead_count,
                fluorescence=fluorescence,
            )
        else:
            return None


@dataclass
class WellSystemLevelMetadata:
    # This class is for data that should be the same across all wells, we just need to grab the data from one well.
    serial_number: str
    controller_version: str
    user: str
    analytical_method: str
    plate_id: str
    regions_of_interest: list[str] = field(default_factory=list)

    @staticmethod
    def create(xml_well: ElementTree.Element) -> WellSystemLevelMetadata:
        serial_number = get_val_from_xml(xml_well, MACHINE_INFO, SERIAL_NUMBER)
        controller_version = get_val_from_xml(
            xml_well, MACHINE_INFO, "MicroControllerVersion"
        )
        user = get_val_from_xml(xml_well, "User")
        analytical_method = get_val_from_xml(xml_well, "RunProtocolDocumentLocation")
        plate_id = get_val_from_xml(xml_well, "PlateID")
        regions = get_element_from_xml(xml_well, RUN_SETTINGS, REGIONS_OF_INTEREST)
        regions_of_interest = []
        for region in regions:
            region_str = str(region.attrib[REGION_NUMBER])
            regions_of_interest.append(region_str)
        return WellSystemLevelMetadata(
            serial_number=serial_number,
            controller_version=controller_version,
            user=user,
            analytical_method=analytical_method,
            plate_id=plate_id,
            regions_of_interest=regions_of_interest,
        )


def validate_xml_structure(full_xml: ElementTree.Element) -> None:
    expected_tags = [
        SAMPLES,
        DOC_LOCATION_TAG,
        DESCRIPTION_TAG,
        PLATE_DIMENSIONS_TAG,
        WELLS_TAG,
    ]
    missing_tags = []
    try:
        present_tags = {child.tag for child in full_xml}
        # Check for missing tags
        for tag in expected_tags:
            if tag not in present_tags:
                missing_tags.append(tag)
    except ElementTree.ParseError as err:
        # Return all expected tags if XML parsing fails
        msg = "Error parsing xml"
        raise AllotropyParserError(msg) from err
    if missing_tags:
        msg = f"Missing expected tags in xml: {missing_tags}"
        raise AllotropeConversionError(msg)


def get_well_name(well_attrib: dict[str, str]) -> str:
    row_name = ROW_NAMES[try_int(well_attrib[ROW_NUMBER], "row_number") - 1]
    column_name = str(well_attrib[COLUMN_NUMBER])
    well_name = row_name + column_name
    return well_name


def map_sample_type(sample_type_tag: str) -> SampleRoleType:
    try:
        sample_type = SAMPLE_ROLE_TYPE_MAPPING[sample_type_tag]
        return sample_type
    except KeyError as err:
        msg = f"{sample_type_tag} is not in the valid list of sample role types: {SAMPLE_ROLE_TYPE_MAPPING.keys()}"
        raise AllotropeConversionError(msg) from err



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import Generic, TypeVar

from allotropy.exceptions import AllotropyParserError

T = TypeVar("T")


class ViewData(Generic[T]):
    def __init__(self, data: dict[str, ViewData[T] | list[T]]):
        self.data = data

    def get_first_key(self) -> str:
        key, *_ = self.data.keys()
        return key

    def iter_keys(self) -> Iterator[list[str]]:
        for key, item in self.data.items():
            if isinstance(item, ViewData):
                for sub_keys in item.iter_keys():
                    yield [key, *sub_keys]
            else:
                yield [key]

    def get_item(self, *keys: str) -> ViewData[T] | list[T]:
        if len(keys) == 0:
            return self.data[self.get_first_key()]

        key, *sub_keys = keys
        item = self.data[key]
        if isinstance(item, ViewData):
            return item.get_item(*sub_keys)
        else:
            return item

    def get_sub_view_data(self, *keys: str) -> ViewData[T]:
        item = self.get_item(*keys)
        if isinstance(item, ViewData):
            return item
        msg = f"Unable to find sub view data with keys: {keys}."
        raise AllotropyParserError(msg)

    def get_leaf_item(self, *keys: str) -> list[T]:
        item = self.get_item(*keys)
        if isinstance(item, ViewData):
            msg = f"Unable to find leaf item of view data with keys: {keys}."
            raise AllotropyParserError(msg)
        return item


class View(ABC, Generic[T]):
    def __init__(self, sub_view: View[T] | None = None):
        self.sub_view = sub_view

    @abstractmethod
    def sort_elements(self, _: list[T]) -> dict[str, list[T]]:
        pass

    def apply(self, elements: list[T]) -> ViewData[T]:
        return ViewData(
            {
                id_: self.sub_view.apply(element) if self.sub_view else element
                for id_, element in self.sort_elements(elements).items()
            }
        )



from dataclasses import fields, MISSING

# We must import this to get all subclasses of TQuantityValue in scope
import allotropy.allotrope.models.shared.definitions.custom  # noqa: F401
from allotropy.allotrope.models.shared.definitions.definitions import TQuantityValue


def _clean_unit(unit: str) -> str:
    return unit.lower().replace("µ", "u")


def _make_unit_entry(quantity_value: type[TQuantityValue]) -> str | None:
    unit_field = next(field for field in fields(quantity_value) if field.name == "unit")
    if unit_field.default == MISSING:
        return None
    return _clean_unit(str(unit_field.default))


UNIT_TO_PROPERTY = {
    _make_unit_entry(quantity_value): quantity_value
    for quantity_value in TQuantityValue.__subclasses__()
}


def get_quantity_class(
    unit: str | None, default: type[TQuantityValue] | None = None
) -> type[TQuantityValue] | None:
    return UNIT_TO_PROPERTY.get(_clean_unit(unit or ""), default)



from __future__ import annotations

from xml.etree import ElementTree as ET  # noqa: N817

from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.electrophoresis.benchling._2024._06.electrophoresis import (
    CalculatedDataItem,
    Data,
    DataSource,
    Error,
    Measurement,
    MeasurementGroup,
    Metadata,
    ProcessedData,
    ProcessedDataFeature,
)
from allotropy.exceptions import (
    AllotropeConversionError,
    AllotropeParsingError,
    get_key_or_error,
    msg_for_error_on_unrecognized_value,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_tapestation_analysis.constants import (
    BRAND_NAME,
    DETECTION_TYPE,
    DEVICE_TYPE,
    NON_CALCULATED_DATA_TAGS_PEAK,
    NON_CALCULATED_DATA_TAGS_REGION,
    NON_CALCULATED_DATA_TAGS_SAMPLE,
    PRODUCT_MANUFACTURER,
    SOFTWARE_NAME,
    UNIT_CLASS_LOOKUP,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import try_float_or_none
from allotropy.parsers.utils.xml import (
    get_element_from_xml,
    get_float_from_xml_or_nan,
    get_float_from_xml_or_none,
    get_val_from_xml,
    get_val_from_xml_or_none,
)


def _get_calculated_data(
    element: ET.Element, excluded_tags: list[str], source_id: str, feature: str
) -> list[CalculatedDataItem]:
    calculated_data = []
    for node in element:
        if (name := node.tag) in excluded_tags:
            continue
        if (value := try_float_or_none(node.text)) is None:
            continue
        calculated_data.append(
            CalculatedDataItem(
                identifier=random_uuid_str(),
                name=name,
                value=value,
                unit=UNITLESS,
                data_sources=[DataSource(identifier=source_id, feature=feature)],
            )
        )

    return calculated_data


def create_metadata(root_element: ET.Element, file_name: str) -> Metadata:
    file_information = get_element_from_xml(root_element, "FileInformation")
    environment = get_element_from_xml(
        root_element, "ScreenTapes/ScreenTape/Environment"
    )
    return Metadata(
        file_name=file_name,
        analyst=get_val_from_xml_or_none(environment, "Experimenter"),
        analytical_method_identifier=get_val_from_xml_or_none(
            file_information, "Assay"
        ),
        data_system_instance_identifier=get_val_from_xml_or_none(
            environment, "Computer"
        ),
        device_identifier=get_val_from_xml_or_none(environment, "InstrumentType"),
        equipment_serial_number=get_val_from_xml_or_none(
            environment, "InstrumentSerialNumber"
        ),
        experimental_data_identifier=get_val_from_xml_or_none(
            file_information, "FileName"
        ),
        # If any, only one of those should appear, so we arbitrarily take the first one
        method_version=get_val_from_xml_or_none(file_information, "RINeVersion")
        or get_val_from_xml_or_none(file_information, "DINVersion"),
        software_version=get_val_from_xml_or_none(environment, "AnalysisVersion"),
        software_name=SOFTWARE_NAME,
        brand_name=BRAND_NAME,
        product_manufacturer=PRODUCT_MANUFACTURER,
        device_type=DEVICE_TYPE,
        detection_type=DETECTION_TYPE,
    )


def _get_description(xml_element: ET.Element) -> str | None:
    comment = get_val_from_xml_or_none(xml_element, "Comment")
    observations = get_val_from_xml_or_none(xml_element, "Observations")
    return f"{comment or ''} {observations or ''}".strip() or None


def _create_peak(peak_element: ET.Element, unit: str) -> ProcessedDataFeature:
    return ProcessedDataFeature(
        identifier=random_uuid_str(),
        height=get_float_from_xml_or_nan(peak_element, "Height"),
        start=get_float_from_xml_or_nan(peak_element, "FromMW"),
        start_unit=unit,
        end=get_float_from_xml_or_nan(peak_element, "ToMW"),
        end_unit=unit,
        position=get_float_from_xml_or_nan(peak_element, "Size"),
        position_unit=unit,
        area=get_float_from_xml_or_nan(peak_element, "Area"),
        relative_area=get_float_from_xml_or_nan(peak_element, "PercentOfTotal"),
        relative_corrected_area=get_float_from_xml_or_nan(
            peak_element, "PercentIntegratedArea"
        ),
        name=get_val_from_xml_or_none(peak_element, "Number"),
        comment=_get_description(peak_element),
    )


def _create_region(
    region_element: ET.Element, region_name: str, unit: str
) -> ProcessedDataFeature:
    return ProcessedDataFeature(
        identifier=random_uuid_str(),
        start=get_float_from_xml_or_nan(region_element, "From"),
        start_unit=unit,
        end=get_float_from_xml_or_nan(region_element, "To"),
        end_unit=unit,
        area=get_float_from_xml_or_nan(region_element, "Area"),
        relative_area=get_float_from_xml_or_nan(region_element, "PercentOfTotal"),
        name=region_name,
        comment=get_val_from_xml_or_none(region_element, "Comment"),
    )


def _create_measurement(
    sample_element: ET.Element, screen_tape: ET.Element, unit: str
) -> tuple[Measurement, list[CalculatedDataItem]]:
    measurement_id = random_uuid_str()
    calculated_data: list[CalculatedDataItem] = []
    calculated_data.extend(
        _get_calculated_data(
            element=sample_element,
            excluded_tags=NON_CALCULATED_DATA_TAGS_SAMPLE,
            source_id=measurement_id,
            feature="sample",
        )
    )
    peaks: list[ProcessedDataFeature] = []
    for peak_element in get_element_from_xml(sample_element, "Peaks").iter("Peak"):
        peaks.append(_create_peak(peak_element, unit))
        calculated_data.extend(
            _get_calculated_data(
                element=peak_element,
                excluded_tags=NON_CALCULATED_DATA_TAGS_PEAK,
                source_id=peaks[-1].identifier,
                feature="peak",
            )
        )

    regions_element = sample_element.find("Regions")
    region_elements = (
        regions_element.iter("Region") if regions_element is not None else []
    )
    regions: list[ProcessedDataFeature] = []
    for idx, region_element in enumerate(
        sorted(
            region_elements,
            key=lambda region: get_val_from_xml(region, "From"),
        ),
        start=1,
    ):
        regions.append(_create_region(region_element, str(idx), unit))
        calculated_data.extend(
            _get_calculated_data(
                element=region_element,
                excluded_tags=NON_CALCULATED_DATA_TAGS_REGION,
                source_id=regions[-1].identifier,
                feature="data region",
            )
        )

    well_number = get_val_from_xml(sample_element, "WellNumber")
    error = (get_val_from_xml_or_none(sample_element, "Alert") or "").strip() or None
    measurement = Measurement(
        identifier=measurement_id,
        measurement_time=get_val_from_xml(screen_tape, "TapeRunDate"),
        compartment_temperature=get_float_from_xml_or_none(
            screen_tape, "ElectrophoresisTemp"
        ),
        location_identifier=well_number,
        sample_identifier=f"{get_val_from_xml(sample_element, 'ScreenTapeID')}_{well_number}",
        description=_get_description(sample_element),
        processed_data=ProcessedData(
            peaks=peaks,
            data_regions=regions,
        ),
        errors=[Error(error=error)] if error else None,
    )
    return measurement, calculated_data


def _get_unit_class(
    root_element: ET.Element,
) -> str:
    peak_unit = get_element_from_xml(
        root_element, "Assay/Units/MolecularWeightUnit"
    ).text
    try:
        peak_unit = peak_unit or ""
        return UNIT_CLASS_LOOKUP[peak_unit]
    except KeyError as e:
        msg = msg_for_error_on_unrecognized_value(
            "Molecular Weight Unit", peak_unit, UNIT_CLASS_LOOKUP.keys()
        )
        raise AllotropeConversionError(msg) from e


def _create_measurement_groups(
    root_element: ET.Element,
) -> tuple[list[MeasurementGroup], list[CalculatedDataItem]]:
    screen_tapes_element = get_element_from_xml(root_element, "ScreenTapes")
    screen_tapes = {
        get_val_from_xml(screen_tape, "ScreenTapeID"): screen_tape
        for screen_tape in screen_tapes_element.iter("ScreenTape")
    }

    measurement_groups: list[MeasurementGroup] = []
    calculated_data: list[CalculatedDataItem] = []
    for sample_element in get_element_from_xml(root_element, "Samples").iter("Sample"):
        screen_tape_id = get_val_from_xml(sample_element, "ScreenTapeID")
        measurement, measurement_calculated_data = _create_measurement(
            sample_element,
            get_key_or_error("ScreenTape ID", screen_tape_id, screen_tapes),
            _get_unit_class(root_element),
        )
        measurement_groups.append(MeasurementGroup(measurements=[measurement]))
        calculated_data.extend(measurement_calculated_data)

    return measurement_groups, calculated_data


def create_data(named_file_contents: NamedFileContents) -> Data:
    try:
        root_element = ET.parse(named_file_contents.contents).getroot()  # noqa: S314
    except ET.ParseError as e:
        msg = f"There was an error when trying to read the xml file: {e}"
        raise AllotropeParsingError(msg) from e

    measurement_groups, calculated_data = _create_measurement_groups(root_element)
    return Data(
        metadata=create_metadata(root_element, named_file_contents.original_file_name),
        measurement_groups=measurement_groups,
        # NOTE: in current implementation, calculated data is reported at global level for some reason.
        # TODO(nstender): should we move this inside of measurements?
        calculated_data=calculated_data,
    )



from __future__ import annotations

from dataclasses import dataclass
import re

import pandas as pd

from allotropy.parsers.utils.uuids import random_uuid_str

# This map is used to coerce the column names coming in the raw data
# into names of the distribution properties.
COLUMN_MAP = {
    "Cumulative Counts/mL": "cumulative_particle_density",
    "Cumulative Count": "cumulative_count",
    "Particle Size(µm)": "particle_size",
    "Differential Counts/mL": "differential_particle_density",
    "Differential Count": "differential_count",
}


VALID_CALCS = ["Average"]


@dataclass(frozen=True, kw_only=True)
class DistributionProperty:
    name: str
    value: float
    distribution_property_id: str


@dataclass(frozen=True, kw_only=True)
class DistributionRow:
    distribution_row_id: str
    properties: list[DistributionProperty]

    def get_property(self, name: str) -> DistributionProperty | None:
        for prop in self.properties:
            if prop.name == name:
                return prop
        return None

    def matches_property_value(
        self, prop_name: str, distribution_row: DistributionRow
    ) -> bool:
        prop = self.get_property(prop_name)
        other_prop = distribution_row.get_property(prop_name)
        if prop and other_prop:
            if prop.value == other_prop.value:
                return True
        return False


@dataclass(frozen=True, kw_only=True)
class Distribution:
    name: str
    data: list[DistributionRow]
    is_calculated: bool
    distribution_id: str

    @staticmethod
    def create(df: pd.DataFrame, name: str, *, is_calculated: bool) -> Distribution:
        data = []
        cols = COLUMN_MAP.values()
        for row in df.index:
            row_data = []
            for col in [x for x in cols if x in df.columns]:
                row_data.append(
                    DistributionProperty(
                        name=col,
                        value=df.loc[row, col],
                        distribution_property_id=random_uuid_str(),
                    )
                )
            data.append(
                DistributionRow(
                    properties=row_data, distribution_row_id=random_uuid_str()
                )
            )
        return Distribution(
            name=name,
            data=data,
            is_calculated=is_calculated,
            distribution_id=random_uuid_str(),
        )


@dataclass(frozen=True, kw_only=True)
class Metadata:
    measurement_time: str
    flush_volume_setting: float
    detector_view_volume: float
    repetition_setting: int
    sample_volume_setting: float
    sample_identifier: str
    dilution_factor_setting: float
    analyst: str
    submitter: str | None = None
    equipment_serial_number: str
    detector_identifier: str
    detector_model_number: str
    software_version: str

    @staticmethod
    def _get_software_version_report_string(report_string: str) -> str:
        match = re.search(r"v(\d+(?:\.\d+)?(?:\.\d+)?)", report_string)
        if match:
            return match.group(1)
        return "Unknown"

    @staticmethod
    def create(df: pd.DataFrame) -> Metadata:
        return Metadata(
            measurement_time=pd.to_datetime(
                str(df.at[8, 5]).replace(".", "-")
            ).isoformat(timespec="microseconds")
            + "Z",
            flush_volume_setting=0,
            detector_view_volume=df.at[9, 5],
            repetition_setting=int(df.at[11, 5]),
            sample_volume_setting=df.at[11, 2],
            sample_identifier=str(df.at[2, 2]),
            dilution_factor_setting=df.at[13, 2],
            analyst=str(df.at[6, 5]),
            submitter=None,
            software_version=Metadata._get_software_version_report_string(df.at[0, 2]),
            equipment_serial_number=str(df.at[4, 5]),
            detector_identifier="",
            detector_model_number=str(df.at[2, 5]),
        )


@dataclass(frozen=True, kw_only=True)
class PharmSpecData:
    metadata: Metadata
    distributions: list[Distribution]

    @staticmethod
    def _get_data_using_key_bounds(
        df: pd.DataFrame, start_key: str, end_key: str
    ) -> pd.DataFrame:
        """Find the data in the raw dataframe. We identify the boundary of the data
        by finding the index first row which contains the word 'Particle' and ending right before
        the index of the first row containing 'Approver'.

        :param df: the raw dataframe
        :param start_key: the key to start the slice
        :parm end_key: the key to end the slice
        :return: the dataframe slice between the stard and end bounds
        """
        start = df[df[1].str.contains(start_key, na=False)].index.values[0]
        end = df[df[0].str.contains(end_key, na=False)].index.values[0] - 1
        return df.loc[start:end, :]

    @staticmethod
    def _extract_data(df: pd.DataFrame) -> pd.DataFrame:
        """Extract the Average data frame from the raw data. Initial use cases have focused on
        only extracting the Average data, not the individual runs. The ASM does support multiple
        Distribution objects, but they don't have names, so it's not possible to pick these out
        after the fact. As such, this extraction only includes the Average data.

        :param df: the raw dataframe
        :return: the average data frame
        """
        data = PharmSpecData._get_data_using_key_bounds(
            df, start_key="Particle", end_key="Approver_"
        )
        data = data.dropna(how="all").dropna(how="all", axis=1)
        data[0] = data[0].ffill()
        data = data.dropna(subset=1).reset_index(drop=True)
        data.columns = pd.Index([x.strip() for x in data.loc[0]])
        data = data.loc[1:, :]
        return data.rename(columns={x: COLUMN_MAP[x] for x in COLUMN_MAP})

    @staticmethod
    def _create_distributions(df: pd.DataFrame) -> list[Distribution]:
        distributions = []
        for g, gdf in df.groupby("Run No."):
            is_calculated = False
            name = str(g)
            if g in VALID_CALCS:
                is_calculated = True
            distribution = Distribution.create(
                df=gdf, name=name, is_calculated=is_calculated
            )
            distributions.append(distribution)
        return distributions

    @staticmethod
    def create(df: pd.DataFrame) -> PharmSpecData:
        data = PharmSpecData._extract_data(df)
        return PharmSpecData(
            metadata=Metadata.create(df),
            distributions=PharmSpecData._create_distributions(data),
        )



from pathlib import Path
import re
import subprocess
import warnings  # noqa: S404, RUF100

from autoflake import fix_file  # type: ignore[import-untyped]
from datamodel_code_generator import (
    DataModelType,
    generate,
    InputFileType,
    PythonVersion,
)

from allotropy.allotrope.schema_parser.backup_manager import (
    backup,
    is_backup_file,
    is_file_changed,
    restore_backup,
)
from allotropy.allotrope.schema_parser.model_class_editor import modify_file
from allotropy.allotrope.schema_parser.path_util import (
    CUSTOM_MODELS_PATH,
    GENERATED_SHARED_PATHS,
    get_model_file_from_schema_path,
    get_rel_schema_path,
    MODEL_DIR_PATH,
    SCHEMA_DIR_PATH,
    UNITS_MODELS_PATH,
)
from allotropy.allotrope.schema_parser.schema_cleaner import SchemaCleaner
from allotropy.allotrope.schema_parser.update_units import update_unit_files


def lint_file(model_path: Path) -> None:
    # The first run of ruff changes typing annotations and causes unused imports. We catch failure
    # due to unused imports.
    try:
        subprocess.check_call(
            f"ruff {model_path} --fix",
            shell=True,  # noqa: S602
            stdout=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError:
        pass
    # The call to autoflake.fix_file removes unused imports.
    fix_file(
        str(model_path),
        {
            "in_place": True,
            "remove_unused_variables": True,
            "write_to_stdout": False,
            "ignore_init_module_imports": False,
            "expand_star_imports": False,
            "remove_all_unused_imports": True,
            "remove_duplicate_keys": True,
            "remove_rhs_for_unused_variables": False,
            "ignore_pass_statements": False,
            "ignore_pass_after_docstring": False,
            "check": False,
            "check_diff": False,
        },
    )
    # The second call to ruff checks for additional rules.
    subprocess.check_call(
        f"ruff {model_path} --fix", shell=True, stdout=subprocess.DEVNULL  # noqa: S602
    )
    subprocess.check_call(
        f"black {model_path}", shell=True, stderr=subprocess.DEVNULL  # noqa: S602
    )


def _generate_schema(model_path: Path, schema_path: Path) -> None:
    with warnings.catch_warnings():
        # We expect duplicated field names for variations with an anyOf schema.
        warnings.filterwarnings(
            "ignore",
            category=UserWarning,
            message="Field name `.*` is duplicated on .*",
        )
        # Known issue - it is OK to treat iri as a string for our purposes.
        warnings.filterwarnings(
            "ignore",
            category=UserWarning,
            message=re.escape(
                "format of 'iri' not understood for 'string' - using default"
            ),
        )
        # Generate models
        generate(
            input_=schema_path,
            output=model_path,
            output_model_type=DataModelType.DataclassesDataclass,
            input_file_type=InputFileType.JsonSchema,
            # Specify base_class as empty when using dataclass
            base_class="",
            target_python_version=PythonVersion.PY_310,
            use_union_operator=True,
        )
        # Import classes from shared files, remove unused classes, format.
        modify_file(model_path, schema_path)
        lint_file(model_path)


def _should_generate_schema(schema_path: Path, schema_regex: str | None = None) -> bool:
    # Skip files in the shared directory
    rel_schema_path = str(get_rel_schema_path(schema_path))
    if rel_schema_path.startswith("shared"):
        return False
    if is_backup_file(rel_schema_path):
        return False
    if schema_regex:
        return bool(re.match(schema_regex, str(rel_schema_path)))
    return True


def make_model_directories(model_path: Path) -> None:
    if model_path == MODEL_DIR_PATH:
        return

    # Call recursively on parents first. We can't just create all directories
    # with a single call, because we want to create __init__ files too.
    make_model_directories(model_path.parent)
    if model_path.exists():
        return
    model_path.mkdir()
    init_path = Path(model_path, "__init__.py")
    if not init_path.exists():
        init_path.touch()


def generate_schemas(
    *,
    dry_run: bool | None = False,
    schema_regex: str | None = None,
) -> list[str]:
    """Generate schemas from JSON schema files.
    :dry_run: If true, does not save changes to any models, but still returns the list of models that would change.
    :schema_regex: If set, filters schemas to generate using regex.
    :return: A list of model files that were changed.
    """
    unit_to_iri: dict[str, str] = {}
    with backup(GENERATED_SHARED_PATHS, restore=dry_run):
        schema_paths = list(Path(SCHEMA_DIR_PATH).rglob("*.json"))
        models_changed = []
        for schema_path in schema_paths:
            if not _should_generate_schema(schema_path, schema_regex):
                continue

            print(  # noqa: T201
                f"Generating models for schema: {get_rel_schema_path(schema_path)}..."
            )
            model_path = Path(
                MODEL_DIR_PATH, get_model_file_from_schema_path(schema_path)
            )
            make_model_directories(model_path.parent)

            with backup(model_path, restore=dry_run), backup(schema_path, restore=True):
                schema_cleaner = SchemaCleaner()
                schema_cleaner.clean_file(str(schema_path))
                unit_to_iri |= schema_cleaner.get_referenced_units()
                _generate_schema(model_path, schema_path)

                if is_file_changed(model_path):
                    models_changed.append(model_path.stem)
                else:
                    restore_backup(model_path)

        update_unit_files(unit_to_iri)
        for path in [UNITS_MODELS_PATH, CUSTOM_MODELS_PATH]:
            lint_file(path)

    return models_changed



from __future__ import annotations

from allotropy.allotrope.models.adm.pcr.benchling._2023._09.dpcr import Model
from allotropy.allotrope.schema_mappers.adm.pcr.BENCHLING._2023._09.dpcr import Mapper
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.qiacuity_dpcr.qiacuity_dpcr_reader import QiacuitydPCRReader
from allotropy.parsers.qiacuity_dpcr.qiacuity_dpcr_structure import create_data
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class QiacuitydPCRParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Qiacuity dPCR"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.RECOMMENDED

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        reader = QiacuitydPCRReader(named_file_contents.contents)
        data = create_data(reader.well_data, named_file_contents.original_file_name)
        return self._get_mapper(Mapper).map_model(data)



#!/usr/bin/env python3
import click

from allotropy.allotrope.schema_parser.reference_resolver import download_schema


@click.command()
@click.argument("schema_url")
def _download_schema(schema_url: str) -> None:
    print(f"Downloading schema at {schema_url}...")
    schema_path = download_schema(schema_url)
    print(f"Downloaded to {schema_path}")


if __name__ == "__main__":
    _download_schema()



from __future__ import annotations

import numpy as np
import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.values import assert_not_none


class MabtechApexContents:
    @staticmethod
    def create(named_file_contents: NamedFileContents) -> MabtechApexContents:
        raw_contents = pd.read_excel(named_file_contents.contents, sheet_name=None)
        return MabtechApexContents(raw_contents)

    def __init__(self, raw_contents: dict[str, pd.DataFrame]) -> None:
        contents = {
            str(name): df.replace(np.nan, None) for name, df in raw_contents.items()
        }

        self.plate_info = self._get_plate_info(contents)
        self.data = self._get_data(contents)

    def _get_plate_info(self, contents: dict[str, pd.DataFrame]) -> SeriesData:
        sheet = assert_not_none(
            contents.get("Plate Information"),
            msg="Unable to find 'Plate Information' sheet.",
        ).dropna(axis=1, how="all")

        data = {}
        for _, * (title, value, *_) in sheet.itertuples():
            if title is None:
                break
            data[str(title)] = None if value is None else str(value)

        return SeriesData(pd.Series(data))

    def _get_data(self, contents: dict[str, pd.DataFrame]) -> pd.DataFrame:
        sheet = assert_not_none(
            contents.get("Plate Database"), msg="Unable to find 'Plate Database' sheet."
        )

        return sheet.dropna(axis=1, how="all")
